# SPDX-License-Identifier: LGPL-2.1-or-later

import datetime
import locale
from collections.abc import Iterable
from pathlib import Path

from mkosi.config import Config
from mkosi.context import Context
from mkosi.curl import curl
from mkosi.distribution import Distribution, debian, join_mirror
from mkosi.installer.apt import AptRepository
from mkosi.log import die
from mkosi.util import startswith


class Installer(debian.Installer, distribution=Distribution.ubuntu):
    @classmethod
    def pretty_name(cls) -> str:
        return "Ubuntu"

    @classmethod
    def default_release(cls) -> str:
        return "devel"

    @classmethod
    def default_tools_tree_distribution(cls) -> Distribution:
        return Distribution.debian

    @classmethod
    def repositories(cls, context: Context, for_image: bool = False) -> Iterable[AptRepository]:
        types = ("deb", "deb-src")
        mirror = None if for_image else context.config.mirror
        snapshot = None if for_image else context.config.snapshot

        components = (
            "main",
            *context.config.repositories,
        )

        if context.config.local_mirror and not for_image:
            yield AptRepository(
                types=("deb",),
                url=context.config.local_mirror,
                suite=context.config.release,
                components=("main",),
                signedby=None,
            )
            return

        if mirror:
            pass
        elif context.config.architecture.is_x86_variant():
            mirror = "http://archive.ubuntu.com/ubuntu"
        else:
            mirror = "http://ports.ubuntu.com"

        signedby = Path("/usr/share/keyrings/ubuntu-archive-keyring.gpg")

        yield AptRepository(
            types=types,
            url=mirror,
            suite=context.config.release,
            components=components,
            signedby=signedby,
            snapshot=snapshot,
        )

        yield AptRepository(
            types=types,
            url=mirror,
            suite=f"{context.config.release}-updates",
            components=components,
            signedby=signedby,
            snapshot=snapshot,
        )

        # Security updates repos are never mirrored. But !x86 are on the ports server.
        if context.config.architecture.is_x86_variant():
            mirror = "http://security.ubuntu.com/ubuntu"
        else:
            mirror = "http://ports.ubuntu.com"

        yield AptRepository(
            types=types,
            url=mirror,
            suite=f"{context.config.release}-security",
            components=components,
            signedby=signedby,
            snapshot=snapshot,
        )

    @classmethod
    def latest_snapshot(cls, config: Config) -> str:
        mirror = config.mirror or "http://snapshot.ubuntu.com"
        release = curl(config, join_mirror(mirror, f"ubuntu/dists/{config.release}-updates/Release"))

        for line in release.splitlines():
            if date := startswith(line, "Date: "):
                # %a and %b parse the abbreviated day of the week and the abbreviated month which are both
                # locale-specific so set the locale to C explicitly to make sure we try to parse the english
                # abbreviations used in the Release file.
                lc = locale.setlocale(locale.LC_TIME)
                try:
                    locale.setlocale(locale.LC_TIME, "C")
                    return datetime.datetime.strptime(date, "%a, %d %b %Y %H:%M:%S %Z").strftime(
                        "%Y%m%dT%H%M%SZ"
                    )
                finally:
                    locale.setlocale(locale.LC_TIME, lc)

        die("Release file is missing Date field")

    @classmethod
    def is_kernel_package(cls, package: str) -> bool:
        return package.startswith("linux-")



# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse
import os
import sys
import tempfile
from pathlib import Path

import mkosi.resources
from mkosi.config import DocFormat
from mkosi.documentation import show_docs
from mkosi.initrd import include_system_config, initrd_common_args, initrd_finalize, process_crypttab
from mkosi.log import ARG_DEBUG, ARG_DEBUG_SHELL, log_setup
from mkosi.run import run, uncaught_exception_handler
from mkosi.util import PathString, resource_path


@uncaught_exception_handler()
def main() -> None:
    log_setup()

    parser = argparse.ArgumentParser(
        prog="mkosi-addon",
        description="Build initrd/cmdline/ucode addon for the current system using mkosi",
        allow_abbrev=False,
        usage="mkosi-addon [options...]",
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="NAME",
        help="Output name",
        default="mkosi-local.addon.efi",
    )

    initrd_common_args(parser)

    args = parser.parse_args()

    if args.show_documentation:
        with resource_path(mkosi.resources) as r:
            show_docs("mkosi-addon", DocFormat.all(), resources=r)
        return

    with tempfile.TemporaryDirectory() as staging_dir:
        cmdline: list[PathString] = [
            "mkosi",
            "--force",
            "--directory", "",
            "--output", args.output,
            "--output-directory", staging_dir,
            "--build-sources", "",
            "--include=mkosi-addon",
            "--extra-tree",
            f"/usr/lib/modules/{args.kernel_version}:/usr/lib/modules/{args.kernel_version}",
            "--extra-tree=/usr/lib/firmware:/usr/lib/firmware",
        ]  # fmt: skip

        if args.debug:
            ARG_DEBUG.set(args.debug)
            cmdline += ["--debug"]
        if args.debug_shell:
            ARG_DEBUG_SHELL.set(args.debug_shell)
            cmdline += ["--debug-shell"]
        if args.debug_sandbox:
            cmdline += ["--debug-sandbox"]

        if os.getuid() == 0:
            cmdline += ["--output-mode=600"]

        cmdline += include_system_config("mkosi-addon")
        cmdline += process_crypttab(Path(staging_dir))

        if Path("/etc/kernel/cmdline").exists():
            cmdline += ["--kernel-command-line", Path("/etc/kernel/cmdline").read_text()]

        run(cmdline, stdin=sys.stdin, stdout=sys.stdout)

        initrd_finalize(Path(staging_dir), args.output, args.output_dir)


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: LGPL-2.1-or-later

import os
import pydoc
from typing import Optional


def page(text: str, enabled: Optional[bool]) -> None:
    if enabled:
        # Initialize less options from $MKOSI_LESS or provide a suitable fallback.
        # F: don't page if one screen
        # X: do not clear screen
        # M: verbose prompt
        # K: quit on ^C
        # R: allow rich formatting
        os.environ["LESS"] = os.getenv("MKOSI_LESS", "FXMKR")
        pydoc.pager(text)
    else:
        print(text)



# SPDX-License-Identifier: LGPL-2.1-or-later

import os
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Optional

from mkosi.log import log_step
from mkosi.run import SandboxProtocol, finalize_passwd_symlinks, nosandbox, run, workdir
from mkosi.sandbox import umask
from mkosi.util import PathString, chdir


def tar_exclude_apivfs_tmp() -> list[str]:
    return [
        "--exclude", "./dev/*",
        "--exclude", "./proc/*",
        "--exclude", "./sys/*",
        "--exclude", "./tmp/*",
        "--exclude", "./run/*",
        "--exclude", "./var/tmp/*",
    ]  # fmt: skip


def make_tar(src: Path, dst: Path, *, sandbox: SandboxProtocol = nosandbox) -> None:
    log_step(f"Creating tar archive {dst}…")

    with dst.open("wb") as f:
        run(
            [
                "tar",
                "--create",
                "--file", "-",
                "--directory", workdir(src, sandbox),
                "--acls",
                "--selinux",
                # --xattrs implies --format=pax
                "--xattrs",
                # PAX format emits additional headers for atime, ctime and mtime
                # that would make the archive non-reproducible.
                "--pax-option=delete=atime,delete=ctime,delete=mtime",
                "--sparse",
                "--force-local",
                *(["--owner=root:0"] if os.getuid() != 0 else []),
                *(["--group=root:0"] if os.getuid() != 0 else []),
                *tar_exclude_apivfs_tmp(),
                ".",
            ],
            stdout=f,
            # Make sure tar uses user/group information from the root directory instead of the host.
            sandbox=sandbox(
                options=[
                    "--ro-bind", src, workdir(src, sandbox),
                    *finalize_passwd_symlinks(workdir(src, sandbox)),
                ],
            ),
        )  # fmt: skip


def can_extract_tar(src: Path) -> bool:
    return ".tar" in src.suffixes[-2:]


def extract_tar(
    src: Path,
    dst: Path,
    *,
    log: bool = True,
    dirs: Sequence[PathString] = (),
    options: Sequence[PathString] = (),
    sandbox: SandboxProtocol = nosandbox,
) -> None:
    if log:
        log_step(f"Extracting tar archive {src}…")

    with umask(~0o755):
        dst.mkdir(exist_ok=True)

    run(
        [
            "tar",
            "--extract",
            "--file", workdir(src, sandbox),
            "--directory", workdir(dst, sandbox),
            "--keep-directory-symlink",
            "--no-overwrite-dir",
            "--same-permissions",
            "--same-owner" if (dst / "etc/passwd").exists() and os.getuid() == 0 else "--numeric-owner",
            "--same-order",
            "--acls",
            "--selinux",
            "--xattrs",
            "--force-local",
            *tar_exclude_apivfs_tmp(),
            *options,
            *dirs,
        ],
        sandbox=sandbox(
            # Make sure tar uses user/group information from the root directory instead of the host.
            options=[
                "--ro-bind", src, workdir(src, sandbox),
                "--bind", dst, workdir(dst, sandbox),
                *finalize_passwd_symlinks(workdir(dst, sandbox)),
            ],
        ),
    )  # fmt: skip


def make_cpio(
    src: Path,
    dst: Path,
    *,
    files: Optional[Iterable[Path]] = None,
    sandbox: SandboxProtocol = nosandbox,
) -> None:
    if not files:
        with chdir(src):
            files = sorted(Path(".").rglob("*"))
    else:
        files = sorted(files)

    log_step(f"Creating cpio archive {dst}…")

    with dst.open("wb") as f:
        run(
            [
                "cpio",
                "--create",
                "--reproducible",
                "--renumber-inodes",
                "--null",
                "--format=newc",
                "--quiet",
                "--directory", workdir(src, sandbox),
                *(["--owner=0:0"] if os.getuid() != 0 else []),
            ],
            input="\0".join(os.fspath(f) for f in files),
            stdout=f,
            sandbox=sandbox(
                options=[
                    "--ro-bind", src, workdir(src, sandbox),
                    *finalize_passwd_symlinks(workdir(src, sandbox))
                ],
            ),
        )  # fmt: skip



# SPDX-License-Identifier: LGPL-2.1-or-later

import tempfile
from collections.abc import Iterable
from pathlib import Path
from typing import Union
from xml.etree import ElementTree

from mkosi.config import Architecture, Config
from mkosi.context import Context
from mkosi.curl import curl
from mkosi.distribution import Distribution, DistributionInstaller, PackageType, join_mirror
from mkosi.installer.dnf import Dnf
from mkosi.installer.rpm import RpmRepository, find_rpm_gpgkey, setup_rpm
from mkosi.installer.zypper import Zypper
from mkosi.log import die
from mkosi.mounts import finalize_certificate_mounts
from mkosi.run import run
from mkosi.versioncomp import GenericVersion


class Installer(DistributionInstaller, distribution=Distribution.opensuse):
    @classmethod
    def pretty_name(cls) -> str:
        return "openSUSE"

    @classmethod
    def filesystem(cls) -> str:
        return "btrfs"

    @classmethod
    def package_type(cls) -> PackageType:
        return PackageType.rpm

    @classmethod
    def default_release(cls) -> str:
        return "tumbleweed"

    @classmethod
    def grub_prefix(cls) -> str:
        return "grub2"

    @classmethod
    def package_manager(cls, config: Config) -> Union[type[Dnf], type[Zypper]]:
        if config.find_binary("zypper"):
            return Zypper
        else:
            return Dnf

    @classmethod
    def setup(cls, context: Context) -> None:
        setup_rpm(context, dbbackend="ndb")
        cls.package_manager(context.config).setup(context, list(cls.repositories(context)))

    @classmethod
    def install(cls, context: Context) -> None:
        packages = ["filesystem"]
        if not any(p.endswith("-release") for p in context.config.packages):
            if context.config.release in ("current", "stable", "leap") or (
                context.config.release != "tumbleweed" and GenericVersion(context.config.release) >= 16
            ):
                packages += ["Leap-release"]
            else:
                packages += ["openSUSE-release"]

        cls.install_packages(context, packages, apivfs=False)

    @classmethod
    def repositories(cls, context: Context) -> Iterable[RpmRepository]:
        if context.config.local_mirror:
            yield RpmRepository(id="local-mirror", url=f"baseurl={context.config.local_mirror}", gpgurls=())
            return

        zypper = cls.package_manager(context.config) is Zypper
        mirror = context.config.mirror or "https://download.opensuse.org"

        if context.config.release == "tumbleweed":
            gpgkeys = tuple(
                p
                for key in ("RPM-GPG-KEY-openSUSE-Tumbleweed", "RPM-GPG-KEY-openSUSE")
                if (p := find_rpm_gpgkey(context, key, required=False))
            )

            if not gpgkeys and not context.config.repository_key_fetch:
                die(
                    "openSUSE GPG keys not found in /usr/share/distribution-gpg-keys",
                    hint="Make sure the distribution-gpg-keys package is installed",
                )

            if zypper and gpgkeys:
                run(
                    [
                        "rpm",
                        "--root=/buildroot",
                        "--import",
                        *(key.removeprefix("file://") for key in gpgkeys),
                    ],
                    sandbox=context.sandbox(
                        options=[
                            *context.rootoptions(),
                            *finalize_certificate_mounts(context.config),
                        ],
                    ),
                )  # fmt: skip

            if context.config.snapshot:
                if context.config.architecture != Architecture.x86_64:
                    die(f"Snapshot= is only supported for x86-64 on {cls.pretty_name()}")

                subdir = f"history/{context.config.snapshot}"
            else:
                if context.config.architecture == Architecture.x86_64:
                    subdir = ""
                elif context.config.architecture == Architecture.arm64:
                    subdir = "ports/aarch64"
                elif context.config.architecture == Architecture.arm:
                    subdir = "ports/armv7hl"
                elif context.config.architecture in (
                    Architecture.ppc64_le,
                    Architecture.ppc64,
                    Architecture.ppc,
                ):
                    subdir = "ports/ppc"
                elif context.config.architecture in (Architecture.s390x, Architecture.s390):
                    subdir = "ports/zsystems"
                elif context.config.architecture == Architecture.riscv64:
                    subdir = "ports/riscv"
                else:
                    die(f"{context.config.architecture} not supported by openSUSE Tumbleweed")

            for repo in ("oss", "non-oss"):
                url = join_mirror(mirror, f"{subdir}/tumbleweed/repo/{repo}")
                yield RpmRepository(
                    id=repo,
                    url=f"baseurl={url}",
                    gpgurls=gpgkeys or (fetch_gpgurls(context, url) if not zypper else ()),
                    enabled=repo == "oss",
                )

                if not context.config.snapshot:
                    for d in ("debug", "source"):
                        if repo == "non-oss" and d == "debug":
                            continue
                        url = join_mirror(mirror, f"{subdir}/{d}/tumbleweed/repo/{repo}")
                        yield RpmRepository(
                            id=f"{repo}-{d}",
                            url=f"baseurl={url}",
                            gpgurls=gpgkeys or (fetch_gpgurls(context, url) if not zypper else ()),
                            enabled=False,
                        )

            if not context.config.snapshot:
                url = join_mirror(mirror, f"{subdir}/update/tumbleweed")
                yield RpmRepository(
                    id="oss-update",
                    url=f"baseurl={url}",
                    gpgurls=gpgkeys or (fetch_gpgurls(context, url) if not zypper else ()),
                )

                url = join_mirror(mirror, f"{subdir}/update/tumbleweed-non-oss")
                yield RpmRepository(
                    id="non-oss-update",
                    url=f"baseurl={url}",
                    gpgurls=gpgkeys or (fetch_gpgurls(context, url) if not zypper else ()),
                    enabled=False,
                )
        else:
            if context.config.snapshot:
                die(f"Snapshot= is only supported for Tumbleweed on {cls.pretty_name()}")

            if (
                context.config.release in ("current", "stable", "leap")
                and context.config.architecture != Architecture.x86_64
            ):
                die(
                    f"{cls.pretty_name()} only supports current and stable releases "
                    "for the x86-64 architecture",
                    hint="Specify either tumbleweed or a specific leap release such as 15.6",
                )

            if context.config.release in ("current", "stable", "leap"):
                release = "openSUSE-current"
            else:
                release = f"leap/{context.config.release}"

            if context.config.architecture == Architecture.x86_64:
                subdir = ""
            elif context.config.architecture == Architecture.arm64:
                subdir = "ports/aarch64"
            elif context.config.architecture == Architecture.arm:
                subdir = "ports/armv7hl"
            elif context.config.architecture in (
                Architecture.ppc64_le,
                Architecture.ppc64,
                Architecture.ppc,
            ):
                subdir = "ports/ppc"
            elif context.config.architecture in (Architecture.s390x, Architecture.s390):
                subdir = "ports/zsystems"
            else:
                die(f"{context.config.architecture} not supported by openSUSE {context.config.release}")

            for repo in ("oss", "non-oss"):
                url = join_mirror(mirror, f"{subdir}/distribution/{release}/repo/{repo}")
                yield RpmRepository(
                    id=repo,
                    url=f"baseurl={url}",
                    gpgurls=fetch_gpgurls(context, url) if not zypper else (),
                    enabled=repo == "oss",
                )

            for d in ("debug", "source"):
                for repo in ("oss", "non-oss"):
                    url = join_mirror(mirror, f"{subdir}/{d}/distribution/{release}/repo/{repo}")
                    yield RpmRepository(
                        id=f"{repo}-{d}",
                        url=f"baseurl={url}",
                        gpgurls=fetch_gpgurls(context, url) if not zypper else (),
                        enabled=False,
                    )

            if (
                context.config.release in ("current", "stable", "leap")
                or GenericVersion(context.config.release) >= 16
            ):
                subdir += f"distribution/{release}/repo"
            else:
                subdir += f"update/{release}"

            for repo in ("oss", "non-oss"):
                url = join_mirror(mirror, f"{subdir}/{repo}")
                yield RpmRepository(
                    id=f"{repo}-update",
                    url=f"baseurl={url}",
                    gpgurls=fetch_gpgurls(context, url) if not zypper else (),
                    enabled=repo == "oss",
                )

    @classmethod
    def architecture(cls, arch: Architecture) -> str:
        a = {
            Architecture.x86_64: "x86_64",
            Architecture.arm64:  "aarch64",
            Architecture.ppc64_le: "ppc64le",
            Architecture.riscv64: "riscv64",
            Architecture.s390x:  "s390x",
        }.get(arch)  # fmt: skip

        if not a:
            die(f"Architecture {a} is not supported by openSUSE")

        return a

    @classmethod
    def latest_snapshot(cls, config: Config) -> str:
        url = join_mirror(config.mirror or "https://download.opensuse.org", "history/latest")
        return curl(config, url).strip()

    @classmethod
    def is_kernel_package(cls, package: str) -> bool:
        return package in ("kernel-default", "kernel-kvmsmall")


def fetch_gpgurls(context: Context, repourl: str) -> tuple[str, ...]:
    gpgurls = [f"{repourl}/repodata/repomd.xml.key"]

    with tempfile.TemporaryDirectory() as d:
        curl(context.config, f"{repourl}/repodata/repomd.xml", output_dir=Path(d))
        xml = (Path(d) / "repomd.xml").read_text()

    root = ElementTree.fromstring(xml)

    tags = root.find("{http://linux.duke.edu/metadata/repo}tags")
    if tags is None:
        die("repomd.xml missing <tags> element")

    for child in tags.iter("{http://linux.duke.edu/metadata/repo}content"):
        if child.text and child.text.startswith("gpg-pubkey"):
            gpgkey = child.text.partition("?")[0]
            gpgurls += [f"{repourl}{gpgkey}"]

    return tuple(gpgurls)



# SPDX-License-Identifier: LGPL-2.1-or-later

from collections.abc import Iterable
from pathlib import Path

from mkosi.config import Architecture
from mkosi.context import Context
from mkosi.distribution import Distribution, debian
from mkosi.installer.apt import AptRepository
from mkosi.log import die


class Installer(debian.Installer, distribution=Distribution.kali):
    @classmethod
    def pretty_name(cls) -> str:
        return "Kali Linux"

    @classmethod
    def default_release(cls) -> str:
        return "kali-rolling"

    @classmethod
    def default_tools_tree_distribution(cls) -> Distribution:
        return Distribution.kali

    @classmethod
    def repositories(cls, context: Context, for_image: bool = False) -> Iterable[AptRepository]:
        mirror = None if for_image else context.config.mirror
        if not mirror:
            mirror = "http://http.kali.org/kali"

        if context.config.snapshot and not for_image:
            die(f"Snapshot= is not supported for {cls.pretty_name()}")

        if context.config.local_mirror and not for_image:
            yield AptRepository(
                types=("deb",),
                url=context.config.local_mirror,
                suite=context.config.release,
                components=("main",),
                signedby=None,
            )
            return

        yield AptRepository(
            types=("deb", "deb-src"),
            url=mirror,
            suite=context.config.release,
            components=("main", *context.config.repositories),
            signedby=Path("/usr/share/keyrings/kali-archive-keyring.gpg"),
        )

    @classmethod
    def architecture(cls, arch: Architecture) -> str:
        a = {
            Architecture.arm64: "arm64",
            Architecture.arm: "armhf",
            Architecture.x86_64: "amd64",
            Architecture.x86: "i386",
        }.get(arch)

        if not a:
            die(f"Architecture {arch} is not supported by {cls.pretty_name()}")

        return a



# SPDX-License-Identifier: LGPL-2.1-or-later

import contextlib
import errno
import logging
import os
import queue
import shlex
import shutil
import signal
import subprocess
import sys
import threading
import uuid
from collections.abc import Awaitable, Collection, Iterator, Mapping, Sequence
from contextlib import AbstractContextManager
from pathlib import Path
from types import TracebackType
from typing import TYPE_CHECKING, Any, Callable, Generic, NoReturn, Optional, Protocol, TypeVar

import mkosi.sandbox
from mkosi.log import ARG_DEBUG, ARG_DEBUG_SANDBOX, ARG_DEBUG_SHELL, die
from mkosi.sandbox import SD_LISTEN_FDS_START, acquire_privileges, joinpath, umask
from mkosi.util import _FILE, PathString, flatten, one_zero, resource_path, unique

# These types are only generic during type checking and not at runtime, leading
# to a TypeError during compilation.
# Let's be as strict as we can with the description for the usage we have.
if TYPE_CHECKING:
    CompletedProcess = subprocess.CompletedProcess[str]
    Popen = subprocess.Popen[str]
else:
    CompletedProcess = subprocess.CompletedProcess
    Popen = subprocess.Popen


T = TypeVar("T")


def ensure_exc_info() -> tuple[type[BaseException], BaseException, TracebackType]:
    exctype, exc, tb = sys.exc_info()
    assert exctype
    assert exc
    assert tb
    return (exctype, exc, tb)


@contextlib.contextmanager
def uncaught_exception_handler(exit: Callable[[int], NoReturn] = sys.exit) -> Iterator[None]:
    rc = 0
    try:
        yield
    except SystemExit as e:
        rc = e.code if isinstance(e.code, int) else 1

        if ARG_DEBUG.get():
            sys.excepthook(*ensure_exc_info())
    except KeyboardInterrupt:
        rc = 1

        if ARG_DEBUG.get():
            sys.excepthook(*ensure_exc_info())
        else:
            logging.error("Interrupted")
    except subprocess.CalledProcessError as e:
        # We always log when subprocess.CalledProcessError is raised, so we don't log again here.
        rc = e.returncode

        # Failures from qemu, ssh and systemd-nspawn are expected and we won't log stacktraces for those.
        # Failures from self come from the forks we spawn to build images in a user namespace. We've already
        # done all the logging for those failures so we don't log stacktraces for those either.
        if (
            ARG_DEBUG.get()
            and e.cmd
            and str(e.cmd[0]) not in ("self", "ssh", "systemd-nspawn")
            and "qemu-system" not in str(e.cmd[0])
        ):
            sys.excepthook(*ensure_exc_info())
    except BaseException:
        sys.excepthook(*ensure_exc_info())
        rc = 1
    finally:
        sys.stdout.flush()
        sys.stderr.flush()
        exit(rc)


def fork_and_wait(target: Callable[..., None], *args: Any, **kwargs: Any) -> None:
    pid = os.fork()
    if pid == 0:
        with uncaught_exception_handler(exit=os._exit):
            target(*args, **kwargs)

    try:
        _, status = os.waitpid(pid, 0)
    except KeyboardInterrupt:
        os.kill(pid, signal.SIGINT)
        _, status = os.waitpid(pid, 0)
    except BaseException:
        os.kill(pid, signal.SIGTERM)
        _, status = os.waitpid(pid, 0)

    rc = os.waitstatus_to_exitcode(status)

    if rc != 0:
        raise subprocess.CalledProcessError(rc, ["self"])


def log_process_failure(sandbox: Sequence[str], cmdline: Sequence[str], returncode: int) -> None:
    if -returncode in (signal.SIGINT, signal.SIGTERM):
        logging.error(f"Interrupted by {signal.Signals(-returncode).name} signal")
    elif returncode < 0:
        logging.error(
            f'"{shlex.join(["mkosi-sandbox", *sandbox, *cmdline] if ARG_DEBUG.get() else cmdline)}"'
            f" was killed by {signal.Signals(-returncode).name} signal."
        )
    elif returncode == 127 and cmdline[0] != "mkosi":
        # Anything invoked beneath /work is a script that we mount into place (so we know it exists). If one
        # of these scripts fails with exit code 127, it's either because the script interpreter was not
        # installed or because one of the commands in the script failed with exit code 127.
        if cmdline[0].startswith("/work"):
            logging.error(f"{cmdline[0]} failed with non-zero exit code 127")
            logging.info(
                "(Maybe a program was not found or the script interpreter (e.g. bash) is not installed?)"
            )
        else:
            logging.error(f"{cmdline[0]} not found.")
    else:
        logging.error(
            f'"{shlex.join(["mkosi-sandbox", *sandbox, *cmdline] if ARG_DEBUG.get() else cmdline)}"'
            f" returned non-zero exit code {returncode}."
        )


def run(
    cmdline: Sequence[PathString],
    check: bool = True,
    stdin: _FILE = None,
    stdout: _FILE = None,
    stderr: _FILE = None,
    input: Optional[str] = None,
    env: Mapping[str, str] = {},
    log: bool = True,
    success_exit_status: Sequence[int] = (0,),
    setup: Sequence[PathString] = (),
    sandbox: AbstractContextManager[Sequence[PathString]] = contextlib.nullcontext([]),
) -> CompletedProcess:
    if input is not None:
        assert stdin is None  # stdin and input cannot be specified together
        stdin = subprocess.PIPE

    with spawn(
        cmdline,
        check=check,
        stdin=stdin,
        stdout=stdout,
        stderr=stderr,
        env=env,
        log=log,
        success_exit_status=success_exit_status,
        setup=setup,
        sandbox=sandbox,
    ) as process:
        out, err = process.communicate(input)

    return CompletedProcess(cmdline, process.returncode, out, err)


@contextlib.contextmanager
def spawn(
    cmdline: Sequence[PathString],
    check: bool = True,
    stdin: _FILE = None,
    stdout: _FILE = None,
    stderr: _FILE = None,
    user: Optional[int] = None,
    group: Optional[int] = None,
    pass_fds: Collection[int] = (),
    env: Mapping[str, str] = {},
    log: bool = True,
    preexec: Optional[Callable[[], None]] = None,
    success_exit_status: Sequence[int] = (0,),
    setup: Sequence[PathString] = (),
    sandbox: AbstractContextManager[Sequence[PathString]] = contextlib.nullcontext([]),
) -> Iterator[Popen]:
    cmd = [os.fspath(x) for x in cmdline]

    if ARG_DEBUG.get():
        logging.info(f"+ {shlex.join(cmd)}")

    if not stdout and not stderr:
        # Unless explicit redirection is done, print all subprocess output on stderr, since we do so as well
        # for mkosi's own output.
        stdout = sys.stderr

    if stdin is None:
        stdin = subprocess.DEVNULL

    env = {
        "PATH": os.environ["PATH"],
        "TERM": os.getenv("TERM", "vt220"),
        "LANG": "C.UTF-8",
        **{k: v for k, v in env.items() if k != "LANG" and not k.startswith("LC_")},
    }

    if "TMPDIR" in os.environ:
        env["TMPDIR"] = os.environ["TMPDIR"]

    for e in ("SYSTEMD_LOG_LEVEL", "SYSTEMD_LOG_LOCATION"):
        if e in os.environ:
            env[e] = os.environ[e]

    if "HOME" not in env:
        env["HOME"] = "/"

    with contextlib.ExitStack() as stack:
        sbx = [os.fspath(x) for x in stack.enter_context(sandbox)]
        apply_sandbox_in_preexec = not setup and not ARG_DEBUG_SANDBOX.get()

        def _preexec() -> None:
            if preexec:
                preexec()

            if sbx and apply_sandbox_in_preexec:
                # The env passed to subprocess.Popen() replaces the environment wholesale so any
                # modifications made by mkosi-sandbox would be overridden if we used that. Hence process the
                # environment in the preexec function.
                os.environ.clear()
                os.environ.update(env)
                try:
                    mkosi.sandbox.main(sbx)
                except Exception:
                    sys.excepthook(*ensure_exc_info())
                    os._exit(1)

        prefix = []
        if sbx and not apply_sandbox_in_preexec:
            module = stack.enter_context(resource_path(sys.modules[__package__ or __name__]))
            prefix = [
                *(["strace", "--detach-on=execve", "--string-limit=256"] if ARG_DEBUG_SANDBOX.get() else []),
                sys.executable, "-SI", os.fspath(module / "sandbox.py"),
                *sbx,
            ]  # fmt: skip

        try:
            proc = subprocess.Popen(
                [*setup, *prefix, *cmdline],
                stdin=stdin,
                stdout=stdout,
                stderr=stderr,
                text=True,
                user=user,
                group=group,
                # Python closes file descriptors after calling the preexec function. Hence we need to tell it
                # to keep the packed file descriptors intact instead of the original ones if --pack-fds is
                # used.
                pass_fds=(
                    range(SD_LISTEN_FDS_START, SD_LISTEN_FDS_START + len(pass_fds))
                    if apply_sandbox_in_preexec and "--pack-fds" in sbx
                    else pass_fds
                ),
                env=env if not sbx or not apply_sandbox_in_preexec else None,
                preexec_fn=_preexec,
            )
        except FileNotFoundError as e:
            die(f"{e.filename} not found.")

        try:
            yield proc
            proc.wait()
        except KeyboardInterrupt:
            proc.send_signal(signal.SIGINT)
            raise
        except BaseException:
            proc.terminate()
            raise
        finally:
            # Make sure any SIGINT/SIGTERM signal we sent is actually processed.
            proc.send_signal(signal.SIGCONT)
            returncode = proc.wait()

        if check and returncode is not None and returncode not in success_exit_status:
            if log:
                log_process_failure(sbx, cmd, returncode)
            if ARG_DEBUG_SHELL.get():
                # --suspend will freeze the debug shell with no way to unfreeze it so strip it from the
                # sandbox if it's there.
                if "--suspend" in prefix:
                    prefix.remove("--suspend")
                if "--suspend" in sbx:
                    sbx.remove("--suspend")
                subprocess.run(
                    [*setup, *prefix, "bash"],
                    check=False,
                    stdin=sys.stdin,
                    text=True,
                    user=user,
                    group=group,
                    env=env if not sbx or not apply_sandbox_in_preexec else None,
                    preexec_fn=_preexec,
                )
            raise subprocess.CalledProcessError(returncode, cmdline)


def finalize_path(
    root: Optional[Path] = None,
    extra: Sequence[Path] = (),
    prefix_usr: bool = False,
    relaxed: bool = False,
) -> str:
    root = root or Path("/")
    path = [os.fspath(p) for p in extra]

    if relaxed:
        path += [
            s
            for s in os.environ["PATH"].split(":")
            if s in ("/usr/bin", "/usr/sbin") or not s.startswith("/usr")
        ]

        # Make sure that /usr/bin and /usr/sbin are always in $PATH.
        path += [s for s in ("/usr/bin", "/usr/sbin") if s not in path]
    else:
        path += ["/usr/bin", "/usr/sbin"]

    if prefix_usr:
        path = [os.fspath(root / s.lstrip("/")) if s in ("/usr/bin", "/usr/sbin") else s for s in path]

    return ":".join(unique(path))


def find_binary(
    *names: PathString,
    root: Optional[Path] = None,
    extra: Sequence[Path] = (),
) -> Optional[Path]:
    root = root or Path("/")
    path = finalize_path(root=root, extra=extra, prefix_usr=True)

    for name in names:
        if any(Path(name).is_relative_to(d) for d in extra):
            pass
        elif Path(name).is_absolute():
            name = root / Path(name).relative_to("/")
        elif "/" in str(name):
            name = root / name

        if binary := shutil.which(name, path=path):
            if root != Path("/") and not Path(binary).is_relative_to(root):
                return Path(binary)
            else:
                return Path("/") / Path(binary).relative_to(root)

    return None


class AsyncioThread(threading.Thread, Generic[T]):
    """
    The default threading.Thread() is not interruptible, so we make our own version by using the concurrency
    feature in python that is interruptible, namely asyncio.

    Additionally, we store any exception that the coroutine raises and re-raise it in join() if no other
    exception was raised before.
    """

    def __init__(
        self, target: Callable[[queue.SimpleQueue[T]], Awaitable[Any]], *args: Any, **kwargs: Any
    ) -> None:
        import asyncio

        self.target = target
        self.loop: queue.SimpleQueue[asyncio.AbstractEventLoop] = queue.SimpleQueue()
        self.exc: queue.SimpleQueue[BaseException] = queue.SimpleQueue()
        self.queue: queue.SimpleQueue[T] = queue.SimpleQueue()
        self.messages: list[T] = []
        super().__init__(*args, **kwargs)

    def run(self) -> None:
        import asyncio

        async def wrapper() -> None:
            self.loop.put(asyncio.get_running_loop())
            await self.target(self.queue)

        try:
            asyncio.run(wrapper())
        except asyncio.CancelledError:
            pass
        except BaseException as e:
            self.exc.put(e)

    def process(self) -> list[T]:
        while not self.queue.empty():
            self.messages += [self.queue.get()]

        return self.messages

    def wait_for(self, expected: T) -> None:
        while (message := self.queue.get()) != expected:
            self.messages += [message]

    def cancel(self) -> None:
        import asyncio.tasks

        loop = self.loop.get()

        for task in asyncio.tasks.all_tasks(loop):
            loop.call_soon_threadsafe(task.cancel)

    def __enter__(self) -> "AsyncioThread[T]":
        self.start()
        return self

    def __exit__(
        self,
        type: Optional[type[BaseException]],
        value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.cancel()
        self.join()
        self.process()

        if type is None:
            try:
                raise self.exc.get_nowait()
            except queue.Empty:
                pass


class SandboxProtocol(Protocol):
    def __call__(
        self,
        *,
        options: Sequence[PathString] = (),
    ) -> AbstractContextManager[list[PathString]]: ...


def nosandbox(
    *,
    options: Sequence[PathString] = (),
) -> AbstractContextManager[list[PathString]]:
    return contextlib.nullcontext([])


def workdir(path: Path, sandbox: Optional[SandboxProtocol] = None) -> str:
    subdir = "/" if sandbox and sandbox == nosandbox else "/work"
    return joinpath(subdir, os.fspath(path))


def finalize_passwd_symlinks(root: PathString) -> list[PathString]:
    """
    If passwd or a related file exists in the apivfs directory, bind mount it over the host files while we
    run the command, to make sure that the command we run uses user/group information from the apivfs
    directory instead of from the host.
    """
    return flatten(
        ("--symlink", Path(root) / "etc" / f, f"/etc/{f}") for f in ("passwd", "group", "shadow", "gshadow")
    )


def network_options(*, network: bool) -> list[PathString]:
    return [
        "--setenv", "SYSTEMD_OFFLINE", one_zero(network),
        *(["--unshare-net"] if not network else []),
    ]  # fmt: skip


@contextlib.contextmanager
def vartmpdir() -> Iterator[Path]:
    # We want to use an empty subdirectory in the host's temporary directory as the sandbox's /var/tmp.
    d = Path(os.getenv("TMPDIR", "/var/tmp")) / f"mkosi-var-tmp-{uuid.uuid4().hex[:16]}"
    d.mkdir()

    try:
        yield d
    finally:
        # A directory that's used as an overlayfs workdir will contain a "work" subdirectory after the
        # overlayfs is unmounted. This "work" subdirectory will have permissions 000 and as such can't be
        # opened or searched unless the user has the CAP_DAC_OVERRIDE capability. shutil.rmtree() will try to
        # search the "work" subdirectory to remove anything in it which will fail with a permission error. To
        # circumvent this, if the work directory exists and is not empty, let's fork off a subprocess where
        # we acquire extra privileges and then invoke shutil.rmtree(). If the work directory exists but is
        # empty, let's just delete the "work" subdirectory first and then invoke shutil.rmtree(). Deleting
        # the subdirectory when it is empty is not a problem because deleting a subdirectory depends on the
        # permissions of the parent directory and not the directory itself.
        try:
            if (p := d / "work").exists():
                p.rmdir()
        except OSError as e:
            if e.errno == errno.ENOTEMPTY:

                def remove() -> None:
                    acquire_privileges()
                    shutil.rmtree(d)

                fork_and_wait(remove)
            else:
                raise
        else:
            shutil.rmtree(d)


@contextlib.contextmanager
def sandbox_cmd(
    *,
    network: bool = False,
    devices: bool = False,
    scripts: Optional[Path] = None,
    tools: Path = Path("/"),
    relaxed: bool = False,
    overlay: Optional[Path] = None,
    options: Sequence[PathString] = (),
    extra: Sequence[Path] = (),
) -> Iterator[list[PathString]]:
    assert not (overlay and relaxed)

    with contextlib.ExitStack() as stack:
        module = stack.enter_context(resource_path(sys.modules[__package__ or __name__]))

        cmdline: list[PathString] = [
            "--proc", "/proc",
            # We mounted a subdirectory of TMPDIR to /var/tmp so we unset TMPDIR so that /tmp or /var/tmp are
            # used instead.
            "--unsetenv", "TMPDIR",
            *network_options(network=network),
        ]  # fmt: skip

        for d in ("usr", "opt"):
            if not (tools / d).exists():
                continue

            if overlay and (overlay / d).exists():
                cmdline += [
                    "--overlay-lowerdir", tools / d,
                    "--overlay-lowerdir", overlay / d,
                    "--overlay", Path("/") / d,
                ]  # fmt: skip
            else:
                cmdline += ["--ro-bind", tools / d, Path("/") / d]

        for d in ("bin", "sbin", "lib", "lib32", "lib64"):
            if (p := tools / d).is_symlink():
                cmdline += ["--symlink", p.readlink(), Path("/") / p.relative_to(tools)]
            elif p.is_dir():
                cmdline += ["--ro-bind", p, Path("/") / p.relative_to(tools)]

        if (tools / "nix/store").exists():
            cmdline += ["--bind", tools / "nix/store", "/nix/store"]

        if relaxed:
            for p in Path("/").iterdir():
                if p not in (
                    Path("/proc"),
                    Path("/usr"),
                    Path("/opt"),
                    Path("/nix"),
                    Path("/bin"),
                    Path("/sbin"),
                    Path("/lib"),
                    Path("/lib32"),
                    Path("/lib64"),
                    Path("/etc"),
                ):
                    if p.is_symlink():
                        cmdline += ["--symlink", p.readlink(), p]
                    else:
                        cmdline += ["--bind", p, p]

            cmdline += ["--ro-bind", tools / "etc", "/etc"]

            if tools != Path("/"):
                for f in ("passwd", "group", "shadow", "gshadow", "nsswitch.conf", "machine-id"):
                    if Path(f"/etc/{f}").exists() and (tools / "etc" / f).exists():
                        cmdline += ["--ro-bind", f"/etc/{f}", f"/etc/{f}"]
        else:
            cmdline += [
                "--dir", "/var/tmp",
                "--dir", "/var/log",
                "--unshare-ipc",
                # apivfs_script_cmd() and chroot_script_cmd() are executed from within the sandbox, but they
                # still use sandbox.py, so we make sure it is available inside the sandbox so it can be
                # executed there as well.
                "--ro-bind", module / "sandbox.py", "/sandbox.py",
            ]  # fmt: skip

            if devices:
                cmdline += ["--bind", "/sys", "/sys", "--bind", "/dev", "/dev"]
            else:
                cmdline += ["--dev", "/dev"]

            # If we're using /usr from a tools tree, we have to use /etc/alternatives and /etc/ld.so.cache
            # from the tools tree as well if they exists since those are directly related to /usr.
            for p in (Path("etc/alternatives"), Path("etc/ld.so.cache")):
                if (tools / p).exists():
                    cmdline += ["--ro-bind", tools / p, Path("/") / p]

            if network and (p := Path("/run/systemd/resolve")).exists():
                cmdline += ["--ro-bind", p, p]

        if network and (p := Path("/etc/resolv.conf")).exists():
            cmdline += ["--ro-bind", p, p]

        path = finalize_path(
            root=tools,
            extra=[Path("/scripts"), *extra] if scripts else extra,
            relaxed=relaxed,
        )
        cmdline += ["--setenv", "PATH", path]

        if scripts:
            cmdline += ["--ro-bind", scripts, "/scripts"]

        tmp: Optional[Path]

        if not overlay and not relaxed:
            tmp = stack.enter_context(vartmpdir())
            yield [
                *cmdline,
                "--bind", tmp, "/var/tmp",
                "--dir", "/etc",
                "--dir", "/var",
                "--dir", "/tmp",
                "--dir", "/run",
                *options,
            ]  # fmt: skip
            return

        if overlay and (overlay / "etc").exists():
            cmdline += ["--ro-bind", overlay / "etc", "/etc"]
        else:
            cmdline += ["--dir", "/etc"]

        for d in ("srv", "media", "mnt", "var", "run", "tmp"):
            tmp = None
            if d not in ("run", "tmp"):
                with umask(~0o755):
                    tmp = stack.enter_context(vartmpdir())

            if overlay and (overlay / d).exists():
                work = None
                if tmp:
                    with umask(~0o755):
                        work = stack.enter_context(vartmpdir())

                cmdline += [
                    "--overlay-lowerdir", overlay / d,
                    "--overlay-upperdir", tmp or "tmpfs",
                    *(["--overlay-workdir", os.fspath(work)] if work else []),
                    "--overlay", Path("/") / d,
                ]  # fmt: skip
            elif not relaxed:
                if tmp:
                    cmdline += ["--bind", tmp, Path("/") / d]
                else:
                    cmdline += ["--dir", Path("/") / d]

        # If we put an overlayfs on /var, and /var/tmp is not in the sandbox tree, make sure /var/tmp is a
        # bind mount of a regular empty directory instead of the overlays so tools like systemd-repart can
        # use the underlying filesystem features from btrfs when using /var/tmp.
        if overlay and not (overlay / "var/tmp").exists():
            tmp = stack.enter_context(vartmpdir())
            cmdline += ["--bind", tmp, "/var/tmp"]

        yield [*cmdline, *options]


def apivfs_options(*, root: Path = Path("/buildroot")) -> list[PathString]:
    return [
        "--tmpfs", root / "run",
        "--tmpfs", root / "tmp",
        "--proc", root / "proc",
        "--dev", root / "dev",
        # Nudge gpg to create its sockets in /run by making sure /run/user/0 exists.
        "--dir", root / "run/user/0",
        # Make sure anything running in the root directory thinks it's in a container. $container can't
        # always be accessed so we write /run/host/container-manager as well which is always accessible.
        "--write", "mkosi", root / "run/host/container-manager",
    ]  # fmt: skip


def chroot_options() -> list[PathString]:
    return [
        # Let's always run as (fake) root when we chroot inside the image as tools executed within the image
        # could have builtin assumptions about files being owned by root.
        "--become-root",
        # Unshare IPC namespace so any tests that exercise IPC related features don't fail with permission
        # errors as --become-root implies unsharing a user namespace which won't have access to the parent's
        # IPC namespace anymore.
        "--unshare-ipc",
        "--setenv", "container", "mkosi",
        "--setenv", "HOME", "/",
        "--setenv", "PATH", "/usr/bin:/usr/sbin",
        "--setenv", "BUILDROOT", "/",
    ]  # fmt: skip


@contextlib.contextmanager
def chroot_cmd(
    *,
    root: Callable[[PathString], list[str]],
    network: bool = False,
    options: Sequence[PathString] = (),
) -> Iterator[list[PathString]]:
    with vartmpdir() as dir:
        cmdline: list[PathString] = [
            *root("/"),
            # We mounted a subdirectory of TMPDIR to /var/tmp so we unset TMPDIR so that /tmp or /var/tmp are
            # used instead.
            "--unsetenv", "TMPDIR",
            *network_options(network=network),
            *apivfs_options(root=Path("/")),
            *chroot_options(),
        ]  # fmt: skip

        if network:
            for p in (Path("/etc/resolv.conf"), Path("/run/systemd/resolve")):
                if p.exists():
                    cmdline += ["--ro-bind", p, p]

        yield [*cmdline, "--bind", dir, "/var/tmp", *options]


def finalize_interpreter(tools: bool) -> str:
    if tools:
        return "python3"

    exe = sys.executable
    if Path(exe).is_relative_to("/usr"):
        return exe

    return "python3"


def glob_in_sandbox(
    *globs: str,
    sandbox: AbstractContextManager[Sequence[PathString]] = contextlib.nullcontext([]),
) -> list[Path]:
    return [
        Path(s)
        for s in run(
            [
                "bash",
                "-c",
                rf"shopt -s nullglob && printf '%s\n' {' '.join(globs)} | xargs -r readlink -f",
            ],
            sandbox=sandbox,
            stdout=subprocess.PIPE,
        )
        .stdout.strip()
        .splitlines()
    ]


def exists_in_sandbox(
    path: PathString,
    sandbox: AbstractContextManager[Sequence[PathString]] = contextlib.nullcontext([]),
) -> bool:
    return (
        run(
            ["bash", "-c", rf"test -e {path}"],
            sandbox=sandbox,
            check=False,
        ).returncode
        == 0
    )



# SPDX-License-Identifier: LGPL-2.1-or-later

import ast
import contextlib
import copy
import enum
import errno
import fcntl
import functools
import hashlib
import importlib
import importlib.resources
import itertools
import logging
import os
import re
import resource
import stat
import tempfile
from collections.abc import Hashable, Iterable, Iterator, Mapping, Sequence
from pathlib import Path
from types import ModuleType
from typing import IO, Any, Callable, Optional, Protocol, TypeVar, Union

from mkosi.log import die
from mkosi.resources import as_file

T = TypeVar("T")
V = TypeVar("V")
S = TypeVar("S", bound=Hashable)

# Borrowed from https://github.com/python/typeshed/blob/3d14016085aed8bcf0cf67e9e5a70790ce1ad8ea/stdlib/3/subprocess.pyi#L24
_FILE = Union[None, int, IO[Any]]
PathString = Union[Path, str]

# Borrowed from
# https://github.com/python/typeshed/blob/ec52bf1adde1d3183d0595d2ba982589df48dff1/stdlib/_typeshed/__init__.pyi#L19
# and
# https://github.com/python/typeshed/blob/ec52bf1adde1d3183d0595d2ba982589df48dff1/stdlib/_typeshed/__init__.pyi#L224
_T_co = TypeVar("_T_co", covariant=True)


class SupportsRead(Protocol[_T_co]):
    def read(self, __length: int = ...) -> _T_co: ...


def dictify(f: Callable[..., Iterator[tuple[T, V]]]) -> Callable[..., dict[T, V]]:
    def wrapper(*args: Any, **kwargs: Any) -> dict[T, V]:
        return dict(f(*args, **kwargs))

    return functools.update_wrapper(wrapper, f)


def tuplify(f: Callable[..., Iterable[T]]) -> Callable[..., tuple[T, ...]]:
    def wrapper(*args: Any, **kwargs: Any) -> tuple[T, ...]:
        return tuple(f(*args, **kwargs))

    return functools.update_wrapper(wrapper, f)


def one_zero(b: bool) -> str:
    return "1" if b else "0"


def is_power_of_2(x: int) -> bool:
    return x > 0 and (x & x - 1 == 0)


def round_up(x: int, blocksize: int = 4096) -> int:
    return (x + blocksize - 1) // blocksize * blocksize


def startswith(s: str, prefix: str) -> Optional[str]:
    if s.startswith(prefix):
        return s.removeprefix(prefix)
    return None


@dictify
def read_env_file(path: PathString) -> Iterator[tuple[str, str]]:
    with Path(path).open() as f:
        for line_number, line in enumerate(f, start=1):
            line = line.rstrip()
            if not line or line.startswith("#"):
                continue
            if m := re.match(r"([a-zA-Z_][a-zA-Z_0-9]+)=(.*)", line):
                name, val = m.groups()
                if val and val[0] in "\"'":
                    val = ast.literal_eval(val)
                yield name, val
            else:
                logging.info(f"{path}:{line_number}: bad line {line!r}")


def format_rlimit(rlimit: int) -> str:
    limits = resource.getrlimit(rlimit)
    soft = "infinity" if limits[0] == resource.RLIM_INFINITY else str(limits[0])
    hard = "infinity" if limits[1] == resource.RLIM_INFINITY else str(limits[1])
    return f"{soft}:{hard}"


def flatten(lists: Iterable[Iterable[T]]) -> list[T]:
    """Flatten a sequence of sequences into a single list."""
    return list(itertools.chain.from_iterable(lists))


@contextlib.contextmanager
def chdir(directory: PathString) -> Iterator[None]:
    old = Path.cwd()

    if old == directory:
        yield
        return

    try:
        os.chdir(directory)
        yield
    finally:
        os.chdir(old)


def make_executable(*paths: Path) -> None:
    for path in paths:
        st = path.stat()
        os.chmod(path, st.st_mode | stat.S_IEXEC)


@contextlib.contextmanager
def flock(path: Path, flags: int = fcntl.LOCK_EX) -> Iterator[int]:
    fd = os.open(path, os.O_CLOEXEC | os.O_RDONLY)
    try:
        fcntl.fcntl(fd, fcntl.FD_CLOEXEC)
        logging.debug(f"Acquiring lock on {path}")
        fcntl.flock(fd, flags)
        logging.debug(f"Acquired lock on {path}")
        yield fd
    finally:
        os.close(fd)


@contextlib.contextmanager
def flock_or_die(path: Path, flags: int = fcntl.LOCK_EX) -> Iterator[Path]:
    try:
        with flock(path, flags | fcntl.LOCK_NB):
            yield path
    except OSError as e:
        if e.errno != errno.EWOULDBLOCK:
            raise e

        die(
            f"Cannot lock {path} as it is locked by another process",
            hint="Maybe another mkosi process is still using it? Use Ephemeral=yes to enable booting "
            "multiple instances of the same image",
        )


@contextlib.contextmanager
def scopedenv(env: Mapping[str, Any]) -> Iterator[None]:
    old = copy.deepcopy(os.environ)
    os.environ |= env

    # python caches the default temporary directory so when we might modify TMPDIR we have to make sure it
    # gets recalculated (see https://docs.python.org/3/library/tempfile.html#tempfile.tempdir).
    tempfile.tempdir = None

    try:
        yield
    finally:
        os.environ = old
        tempfile.tempdir = None


class StrEnum(enum.Enum):
    def __str__(self) -> str:
        assert isinstance(self.value, str)
        return self.value

    # Used by enum.auto() to get the next value.
    @staticmethod
    def _generate_next_value_(name: str, start: int, count: int, last_values: Sequence[str]) -> str:
        return name.replace("_", "-")

    @classmethod
    def values(cls) -> list[str]:
        return list(s.replace("_", "-") for s in map(str, cls.__members__))

    @classmethod
    def choices(cls) -> list[str]:
        return [*cls.values(), ""]


def parents_below(path: Path, below: Path) -> list[Path]:
    parents = list(path.parents)
    return parents[: parents.index(below)]


@contextlib.contextmanager
def resource_path(mod: ModuleType) -> Iterator[Path]:
    t = importlib.resources.files(mod)
    with as_file(t) as p:
        # Make sure any temporary directory that the resources are unpacked in is accessible to the invoking
        # user so that any commands executed as the invoking user can access files within it.
        if (
            p.parent.parent == Path(os.getenv("TMPDIR", "/tmp"))
            and stat.S_IMODE(p.parent.stat().st_mode) == 0o700
        ):
            p.parent.chmod(0o755)

        yield p


def hash_file(path: Path) -> str:
    # TODO Replace with hashlib.file_digest after dropping support for Python 3.10.
    h = hashlib.sha256()
    b = bytearray(16 * 1024**2)
    mv = memoryview(b)

    with path.open("rb", buffering=0) as f:
        while n := f.readinto(mv):
            h.update(mv[:n])

    return h.hexdigest()


def try_or(fn: Callable[..., T], exception: type[Exception], default: T) -> T:
    try:
        return fn()
    except exception:
        return default


def groupby(seq: Sequence[T], key: Callable[[T], S]) -> list[tuple[S, list[T]]]:
    grouped: dict[S, list[T]] = {}

    for i in seq:
        k = key(i)

        if k not in grouped:
            grouped[k] = []

        grouped[k].append(i)

    return [(key, group) for key, group in grouped.items()]


def unique(seq: Sequence[T]) -> list[T]:
    return list(dict.fromkeys(seq))


def mandatory_variable(name: str) -> str:
    try:
        return os.environ[name]
    except KeyError:
        die(f"${name} must be set in the environment")



# SPDX-License-Identifier: LGPL-2.1-or-later

import dataclasses
import json
import subprocess
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any, Final, Optional

from mkosi.log import die
from mkosi.run import SandboxProtocol, nosandbox, run, workdir


@dataclasses.dataclass(frozen=True)
class Partition:
    type: str
    uuid: str
    partno: Optional[int]
    split_path: Optional[Path]
    roothash: Optional[str]

    @classmethod
    def from_dict(cls, dict: Mapping[str, Any]) -> "Partition":
        return cls(
            type=dict["type"],
            uuid=dict["uuid"],
            partno=int(partno) if (partno := dict.get("partno")) else None,
            # We have to translate the sandbox path to the path on the host by removing the /work prefix.
            split_path=(
                Path(p.removeprefix("/work")) if ((p := dict.get("split_path")) and p != "-") else None
            ),
            roothash=dict.get("roothash"),
        )

    GRUB_BOOT_PARTITION_UUID: Final[str] = "21686148-6449-6e6f-744e-656564454649"


def find_partitions(image: Path, *, sandbox: SandboxProtocol = nosandbox) -> list[Partition]:
    output = json.loads(
        run(
            ["systemd-repart", "--json=short", workdir(image, sandbox)],
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            sandbox=sandbox(options=["--ro-bind", image, workdir(image, sandbox)]),
        ).stdout
    )
    return [Partition.from_dict(d) for d in output]


def finalize_roothash(partitions: Sequence[Partition]) -> Optional[str]:
    roothash: Optional[str] = None
    usrhash: Optional[str] = None

    for p in partitions:
        if (h := p.roothash) is None:
            continue

        if not (p.type.startswith("usr") or p.type.startswith("root")):
            die(f"Found roothash property on unexpected partition type {p.type}")

        # When the partition is deferred/skipped the roothash is listed as the literal 'TBD' string
        if h == "TBD":
            continue

        # When there's multiple verity enabled root or usr partitions, the first one wins.
        if p.type.startswith("usr"):
            usrhash = usrhash or h
        else:
            roothash = roothash or h

    return f"roothash={roothash}" if roothash else f"usrhash={usrhash}" if usrhash else None


def finalize_root(partitions: Sequence[Partition]) -> Optional[str]:
    root = finalize_roothash(partitions)
    if not root:
        root = next((f"root=PARTUUID={p.uuid}" for p in partitions if p.type.startswith("root")), None)
    if not root:
        root = next((f"mount.usr=PARTUUID={p.uuid}" for p in partitions if p.type.startswith("usr")), None)

    return root



# SPDX-License-Identifier: LGPL-2.1-or-later

import functools
import itertools
import string
from typing import Final


@functools.total_ordering
class GenericVersion:
    # These constants follow the convention of the return value of rpmdev-vercmp that are followed
    # by systemd-analyze compare-versions when called with only two arguments (without a comparison
    # operator), recreated in the compare_versions method.
    _EQUAL: Final[int] = 0
    _RIGHT_SMALLER: Final[int] = 1
    _LEFT_SMALLER: Final[int] = -1

    def __init__(self, version: str):
        self._version = version

    @classmethod
    def compare_versions(cls, v1: str, v2: str) -> int:
        """Implements comparison according to UAPI Group Version Format Specification"""

        def rstrip_invalid_version_chars(s: str) -> str:
            valid_version_chars = {*string.ascii_letters, *string.digits, "~", "-", "^", "."}
            for i, c in enumerate(s):
                if c in valid_version_chars:
                    return s[i:]
            return ""

        def digit_prefix(s: str) -> str:
            return "".join(itertools.takewhile(lambda c: c in string.digits, s))

        def letter_prefix(s: str) -> str:
            return "".join(itertools.takewhile(lambda c: c in string.ascii_letters, s))

        while True:
            # Any characters which are outside of the set of listed above (a-z, A-Z, 0-9, -, ., ~,
            # ^) are skipped in both strings. In particular, this means that non-ASCII characters
            # that are Unicode digits or letters are skipped too.
            v1 = rstrip_invalid_version_chars(v1)
            v2 = rstrip_invalid_version_chars(v2)

            # If the remaining part of one of strings starts with "~": if other remaining part does
            # not start with ~, the string with ~ compares lower. Otherwise, both tilde characters
            # are skipped.

            if v1.startswith("~") and v2.startswith("~"):
                v1 = v1.removeprefix("~")
                v2 = v2.removeprefix("~")
            elif v1.startswith("~"):
                return cls._LEFT_SMALLER
            elif v2.startswith("~"):
                return cls._RIGHT_SMALLER

            # If one of the strings has ended: if the other string hasn’t, the string that has
            # remaining characters compares higher. Otherwise, the strings compare equal.

            if not v1 and not v2:
                return cls._EQUAL
            elif not v1 and v2:
                return cls._LEFT_SMALLER
            elif v1 and not v2:
                return cls._RIGHT_SMALLER

            # If the remaining part of one of strings starts with "-": if the other remaining part
            # does not start with -, the string with - compares lower. Otherwise, both minus
            # characters are skipped.

            if v1.startswith("-") and v2.startswith("-"):
                v1 = v1.removeprefix("-")
                v2 = v2.removeprefix("-")
            elif v1.startswith("-"):
                return cls._LEFT_SMALLER
            elif v2.startswith("-"):
                return cls._RIGHT_SMALLER

            # If the remaining part of one of strings starts with "^": if the other remaining part
            # does not start with ^, the string with ^ compares higher. Otherwise, both caret
            # characters are skipped.

            if v1.startswith("^") and v2.startswith("^"):
                v1 = v1.removeprefix("^")
                v2 = v2.removeprefix("^")
            elif v1.startswith("^"):
                # TODO: bug?
                return cls._LEFT_SMALLER  # cls._RIGHT_SMALLER
            elif v2.startswith("^"):
                return cls._RIGHT_SMALLER  # cls._LEFT_SMALLER

            # If the remaining part of one of strings starts with ".": if the other remaining part
            # does not start with ., the string with . compares lower. Otherwise, both dot
            # characters are skipped.

            if v1.startswith(".") and v2.startswith("."):
                v1 = v1.removeprefix(".")
                v2 = v2.removeprefix(".")
            elif v1.startswith("."):
                return cls._LEFT_SMALLER
            elif v2.startswith("."):
                return cls._RIGHT_SMALLER

            # If either of the remaining parts starts with a digit: numerical prefixes are compared
            # numerically. Any leading zeroes are skipped. The numerical prefixes (until the first
            # non-digit character) are evaluated as numbers. If one of the prefixes is empty, it
            # evaluates as 0. If the numbers are different, the string with the bigger number
            # compares higher. Otherwise, the comparison continues at the following characters at
            # point 1.

            v1_digit_prefix = digit_prefix(v1)
            v2_digit_prefix = digit_prefix(v2)

            if v1_digit_prefix or v2_digit_prefix:
                v1_digits = int(v1_digit_prefix) if v1_digit_prefix else 0
                v2_digits = int(v2_digit_prefix) if v2_digit_prefix else 0

                if v1_digits < v2_digits:
                    return cls._LEFT_SMALLER
                elif v1_digits > v2_digits:
                    return cls._RIGHT_SMALLER

                v1 = v1.removeprefix(v1_digit_prefix)
                v2 = v2.removeprefix(v2_digit_prefix)
                continue

            # Leading alphabetical prefixes are compared alphabetically. The substrings are
            # compared letter-by-letter. If both letters are the same, the comparison continues
            # with the next letter. Capital letters compare lower than lower-case letters (A <
            # a). When the end of one substring has been reached (a non-letter character or the end
            # of the whole string), if the other substring has remaining letters, it compares
            # higher. Otherwise, the comparison continues at the following characters at point 1.

            v1_letter_prefix = letter_prefix(v1)
            v2_letter_prefix = letter_prefix(v2)

            if v1_letter_prefix < v2_letter_prefix:
                return cls._LEFT_SMALLER
            elif v1_letter_prefix > v2_letter_prefix:
                return cls._RIGHT_SMALLER

            v1 = v1.removeprefix(v1_letter_prefix)
            v2 = v2.removeprefix(v2_letter_prefix)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, (str, int)):
            other = GenericVersion(str(other))
        elif not isinstance(other, GenericVersion):
            return False
        return self.compare_versions(self._version, other._version) == self._EQUAL

    def __lt__(self, other: object) -> bool:
        if isinstance(other, (str, int)):
            other = GenericVersion(str(other))
        elif not isinstance(other, GenericVersion):
            return False
        return self.compare_versions(self._version, other._version) == self._LEFT_SMALLER

    def __str__(self) -> str:
        return self._version

    def __repr__(self) -> str:
        return f"GenericVersion('{self._version}')"



# SPDX-License-Identifier: LGPL-2.1-or-later

import dataclasses
import textwrap
from collections.abc import Sequence
from pathlib import Path
from typing import Final, Optional

from mkosi.config import Config, ConfigFeature
from mkosi.context import Context
from mkosi.installer import PackageManager
from mkosi.log import die
from mkosi.run import CompletedProcess, run, workdir
from mkosi.sandbox import umask
from mkosi.util import _FILE, PathString


@dataclasses.dataclass(frozen=True)
class AptRepository:
    types: tuple[str, ...]
    url: str
    suite: str
    components: tuple[str, ...]
    signedby: Optional[Path]
    snapshot: Optional[str] = None

    def __str__(self) -> str:
        return textwrap.dedent(
            f"""\
            Types: {" ".join(self.types)}
            URIs: {self.url}
            Suites: {self.suite}
            Components: {" ".join(self.components)}
            {"Signed-By" if self.signedby else "Trusted"}: {self.signedby or "yes"}
            {f"Snapshot: {self.snapshot}" if self.snapshot else ""}

            """
        )


class Apt(PackageManager):
    documentation_exclude_globs: Final[list[str]] = [
        "usr/share/doc/*",
        "usr/share/man/*",
        "usr/share/groff/*",
        "usr/share/gtk-doc/*",
        "usr/share/info/*",
    ]

    @classmethod
    def executable(cls, config: Config) -> str:
        return "apt-get"

    @classmethod
    def subdir(cls, config: Config) -> Path:
        return Path("apt")

    @classmethod
    def package_subdirs(cls, cache: Path) -> list[tuple[Path, Path]]:
        return [(Path("archives"), Path("archives"))]

    @classmethod
    def package_globs(cls) -> list[str]:
        return ["*.deb", "*.ddeb"]

    @classmethod
    def state_subdirs(cls) -> list[Path]:
        return [Path("lists")]

    @classmethod
    def dpkg_cmd(cls, command: str) -> list[PathString]:
        return [
            command,
            "--admindir=/buildroot/var/lib/dpkg",
            "--root=/buildroot",
        ]

    @classmethod
    def scripts(cls, context: Context) -> dict[str, list[PathString]]:
        cmd = cls.apivfs_script_cmd(context)

        return {
            **{
                command: cmd + cls.env_cmd(context) + cls.cmd(context, command)
                for command in (
                    "apt",
                    "apt-cache",
                    "apt-cdrom",
                    "apt-config",
                    "apt-extracttemplates",
                    "apt-get",
                    "apt-key",
                    "apt-mark",
                    "apt-sortpkgs",
                )
            },
            **{
                command: cmd + cls.dpkg_cmd(command)
                for command in (
                    "dpkg",
                    "dpkg-query",
                )
            },
            "mkosi-install":   ["apt-get", "install"],
            "mkosi-upgrade":   ["apt-get", "upgrade"],
            "mkosi-remove":    ["apt-get", "purge"],
            "mkosi-reinstall": ["apt-get", "install", "--reinstall"],
        }  # fmt: skip

    @classmethod
    def options(cls, *, root: PathString, apivfs: bool = True) -> list[PathString]:
        return super().options(root=root, apivfs=apivfs) + ["--dir", "/var/lib/apt/lists/partial"]

    @classmethod
    def setup(cls, context: Context, repositories: Sequence[AptRepository]) -> None:
        (context.sandbox_tree / "etc/apt").mkdir(exist_ok=True, parents=True)
        (context.sandbox_tree / "etc/apt/apt.conf.d").mkdir(exist_ok=True, parents=True)
        (context.sandbox_tree / "etc/apt/preferences.d").mkdir(exist_ok=True, parents=True)
        (context.sandbox_tree / "etc/apt/sources.list.d").mkdir(exist_ok=True, parents=True)

        # We have a special apt.conf outside of the sandbox tree that only configures "Dir::Etc" that we pass
        # to APT_CONFIG to tell apt it should read config files from /etc/apt in case this is overridden by
        # distributions.  This is required because apt parses CLI configuration options after parsing its
        # configuration files and as such we can't use CLI options to tell apt where to look for
        # configuration files.
        config = context.sandbox_tree / "etc/apt.conf"
        if not config.exists():
            config.write_text(
                textwrap.dedent(
                    """\
                    Dir::Etc "/etc/apt";
                    """
                )
            )

        sources = context.sandbox_tree / "etc/apt/sources.list.d/mkosi.sources"
        if not sources.exists():
            for repo in repositories:
                if repo.signedby and not (context.config.tools() / str(repo.signedby).lstrip("/")).exists():
                    die(
                        f"Keyring for repo {repo.url} not found at {repo.signedby}",
                        hint="Make sure the right keyring package (e.g. debian-archive-keyring, "
                        "kali-archive-keyring or ubuntu-keyring) is installed",
                    )

            with sources.open("w") as f:
                for repo in repositories:
                    f.write(str(repo))

    @classmethod
    def finalize_environment(cls, context: Context) -> dict[str, str]:
        env = {
            "APT_CONFIG": "/etc/apt.conf",
            "DEBIAN_FRONTEND": "noninteractive",
            "DEBCONF_INTERACTIVE_SEEN": "true",
        }

        if (
            "INITRD" not in context.config.finalize_environment()
            and context.config.bootable != ConfigFeature.disabled
        ):
            env["INITRD"] = "No"

        return super().finalize_environment(context) | env

    @classmethod
    def cmd(cls, context: Context, command: str = "apt-get") -> list[PathString]:
        cmdline: list[PathString] = [
            command,
            "-o", f"APT::Architecture={cls.architecture(context)}",
            "-o", f"APT::Architectures={cls.architecture(context)}",
            "-o", f"APT::Install-Recommends={str(context.config.with_recommends).lower()}",
            "-o", "APT::Immediate-Configure=off",
            "-o", "APT::Get::Assume-Yes=true",
            "-o", "APT::Get::AutomaticRemove=true",
            "-o", "APT::Get::Allow-Change-Held-Packages=true",
            "-o", "APT::Get::Allow-Remove-Essential=true",
            "-o", "APT::Sandbox::User=root",
            "-o", "Acquire::AllowReleaseInfoChange=true",
            "-o", "Acquire::Check-Valid-Until=false",
            "-o", "Dir::Cache=/var/cache/apt",
            "-o", "Dir::State=/buildroot/var/lib/apt",
            "-o", "Dir::State::lists=/var/lib/apt/lists/",
            "-o", "Dir::Log=/var/log/apt",
            "-o", "Dir::State::Status=/buildroot/var/lib/dpkg/status",
            "-o", f"Dir::Bin::DPkg={context.config.find_binary('dpkg')}",
            "-o", "Debug::NoLocking=true",
            "-o", "DPkg::Options::=--root=/buildroot",
            "-o", "DPkg::Options::=--force-unsafe-io",
            "-o", "DPkg::Options::=--force-architecture",
            "-o", "DPkg::Options::=--force-depends",
            "-o", "DPkg::Options::=--no-debsig",
            "-o", "DPkg::Use-Pty=false",
            "-o", "DPkg::Install::Recursive::Minimum=1000",
            "-o", "pkgCacheGen::ForceEssential=,",
        ]  # fmt: skip

        if not context.config.repository_key_check:
            cmdline += [
                "-o", "Acquire::AllowInsecureRepositories=true",
                "-o", "Acquire::AllowDowngradeToInsecureRepositories=true",
                "-o", "APT::Get::AllowUnauthenticated=true",
            ]  # fmt: skip

        if not context.config.with_docs:
            cmdline += [
                f"--option=DPkg::Options::=--path-exclude=/{glob}"
                for glob in cls.documentation_exclude_globs
            ]
            cmdline += ["--option=DPkg::Options::=--path-include=/usr/share/doc/*/copyright"]

        if context.config.proxy_url:
            cmdline += [
                "-o", f"Acquire::http::Proxy={context.config.proxy_url}",
                "-o", f"Acquire::https::Proxy={context.config.proxy_url}",
            ]  # fmt: skip

        return cmdline

    @classmethod
    def invoke(
        cls,
        context: Context,
        operation: str,
        arguments: Sequence[str] = (),
        *,
        apivfs: bool = False,
        options: Sequence[PathString] = (),
        stdout: _FILE = None,
    ) -> CompletedProcess:
        with umask(~0o755):
            # TODO: Drop once apt 2.5.4 is widely available.
            (context.root / "var/lib/dpkg").mkdir(parents=True, exist_ok=True)
            (context.root / "var/lib/dpkg/status").touch(exist_ok=True)

            (context.root / "var/lib/dpkg/available").touch(exist_ok=True)

        return run(
            cls.cmd(context) + [operation, *arguments],
            sandbox=cls.sandbox(context, apivfs=apivfs, options=options),
            env=cls.finalize_environment(context),
            stdout=stdout,
        )

    @classmethod
    def install(
        cls,
        context: Context,
        packages: Sequence[str],
        *,
        apivfs: bool = True,
        allow_downgrade: bool = False,
    ) -> None:
        # Debian policy is to start daemons by default. The policy-rc.d script can be used choose which ones
        # to start. Let's install one that denies all daemon startups.
        # See https://people.debian.org/~hmh/invokerc.d-policyrc.d-specification.txt for more information.
        # Note: despite writing in /usr/sbin, this file is not shipped by the OS and instead should be
        # managed by the admin.
        policyrcd = context.root / "usr/sbin/policy-rc.d"
        with umask(~0o755):
            policyrcd.parent.mkdir(parents=True, exist_ok=True)
        with umask(~0o644):
            policyrcd.write_text("#!/bin/sh\nexit 101\n")

        arguments = []

        if allow_downgrade:
            arguments += ["--allow-downgrades"]

        arguments += [*packages]

        cls.invoke(context, "install", arguments, apivfs=apivfs)

        policyrcd.unlink()

        # systemd-gpt-auto-generator is disabled by default in Ubuntu:
        # https://git.launchpad.net/ubuntu/+source/systemd/tree/debian/systemd.links?h=ubuntu/noble-proposed.
        # Let's make sure it is enabled by default in our images.
        (context.root / "etc/systemd/system-generators/systemd-gpt-auto-generator").unlink(missing_ok=True)

    @classmethod
    def remove(cls, context: Context, packages: Sequence[str]) -> None:
        cls.invoke(context, "purge", packages, apivfs=True)

    @classmethod
    def sync(cls, context: Context, force: bool) -> None:
        cls.invoke(context, "update")

    @classmethod
    def createrepo(cls, context: Context) -> None:
        names = [d.name for glob in cls.package_globs() for d in context.repository.glob(glob)]
        if not names:
            return

        with (context.repository / "Packages").open("wb") as f:
            run(
                ["apt-ftparchive", "packages", "."],
                stdout=f,
                sandbox=context.sandbox(
                    options=[
                        "--ro-bind", context.repository, workdir(context.repository),
                        "--chdir", workdir(context.repository),
                    ],
                ),
            )  # fmt: skip

        (context.sandbox_tree / "etc/apt/sources.list.d").mkdir(parents=True, exist_ok=True)
        (context.sandbox_tree / "etc/apt/sources.list.d/mkosi-local.sources").write_text(
            textwrap.dedent(
                """\
                Enabled: yes
                Types: deb
                URIs: file:///repository
                Suites: ./
                Trusted: yes
                """
            )
        )

        (context.sandbox_tree / "etc/apt/preferences.d").mkdir(parents=True, exist_ok=True)
        (context.sandbox_tree / "etc/apt/preferences.d/mkosi-local.pref").write_text(
            textwrap.dedent(
                """\
                Package: *
                Pin: origin ""
                Pin-Priority: 1100
                """
            )
        )

        cls.invoke(
            context,
            "update",
            arguments=[
                "-o", "Dir::Etc::sourcelist=sources.list.d/mkosi-local.sources",
                "-o", "Dir::Etc::sourceparts=-",
                "-o", "APT::Get::List-Cleanup=0",
            ],
        )  # fmt: skip



# SPDX-License-Identifier: LGPL-2.1-or-later

import shutil
from collections.abc import Iterable

from mkosi.config import Architecture, Config
from mkosi.context import Context
from mkosi.distribution import Distribution, DistributionInstaller, PackageType
from mkosi.installer import PackageManager
from mkosi.installer.apk import Apk, ApkRepository
from mkosi.log import complete_step, die


class Installer(DistributionInstaller, distribution=Distribution.postmarketos):
    @classmethod
    def pretty_name(cls) -> str:
        return "postmarketOS"

    @classmethod
    def filesystem(cls) -> str:
        return "ext4"

    @classmethod
    def package_type(cls) -> PackageType:
        return PackageType.apk

    @classmethod
    def default_release(cls) -> str:
        return "edge"

    @classmethod
    def default_tools_tree_distribution(cls) -> Distribution:
        return Distribution.postmarketos

    @classmethod
    def package_manager(cls, config: Config) -> type[PackageManager]:
        return Apk

    @classmethod
    def setup(cls, context: Context) -> None:
        with complete_step("Setting up postmarketOS keyring"):
            keys = context.sandbox_tree / "etc/apk/keys"
            keys.mkdir(parents=True, exist_ok=True)

            for d in [
                context.config.tools() / "usr/lib/apk/keys",
                context.config.tools() / "usr/share/distribution-gpg-keys/alpine-linux",
                context.config.tools() / "usr/share/distribution-gpg-keys/postmarketos",
            ]:
                if not d.exists():
                    continue

                # Do not overwrite keys in /etc/apk/keys to make sure that user provided keys take priority.
                for key in d.iterdir():
                    if key.is_dir():
                        continue

                    dest = keys / key.name
                    if dest.exists():
                        continue

                    shutil.copy2(key, dest)

        Apk.setup(context, list(cls.repositories(context)))

    @classmethod
    def install(cls, context: Context) -> None:
        # TODO: Create merged /usr manually for now until our upstream (Alpine Linux) supports it:
        # https://gitlab.alpinelinux.org/alpine/aports/-/merge_requests/85504
        for dir in ["lib", "bin", "sbin"]:
            (context.root / "usr" / dir).mkdir(parents=True, exist_ok=True)
            (context.root / dir).symlink_to(f"usr/{dir}")

        cls.install_packages(context, ["postmarketos-baselayout", "postmarketos-release"], apivfs=False)

    @classmethod
    def repositories(cls, context: Context) -> Iterable[ApkRepository]:
        if context.config.release != "edge":
            die(f"Only 'edge' release is currently supported, got '{context.config.release}'")

        if context.config.local_mirror:
            yield ApkRepository(url=context.config.local_mirror)
            return

        # Alpine repos
        # Note: "testing" is enabled here because it's also enabled by default when pmbootstrap builds pmOS
        # images, sometimes pmOS pkgs temporarily depend on things in testing.
        for repo_name in ["main", "community", "testing"]:
            yield ApkRepository(
                url=f"https://dl-cdn.alpinelinux.org/alpine/{context.config.release}/{repo_name}"
            )

        # postmarketOS repos
        mirror = context.config.mirror or "https://mirror.postmarketos.org/postmarketos"
        subdir = "master" if context.config.release == "edge" else f"v{context.config.release}"

        yield ApkRepository(url=f"{mirror}/extra-repos/systemd/{subdir}")
        yield ApkRepository(url=f"{mirror}/{subdir}")

    @classmethod
    def architecture(cls, arch: Architecture) -> str:
        a = {
            Architecture.x86_64: "x86_64",
            Architecture.arm64:  "aarch64",
            Architecture.arm:    "armv7",
        }.get(arch)  # fmt: skip

        if not a:
            die(f"Architecture {a} is not supported by postmarketOS")

        return a

    @classmethod
    def is_kernel_package(cls, package: str) -> bool:
        if not package.startswith("linux-"):
            return False

        if package.endswith(("-doc", "-dev", "-manual")):
            return False

        # These pkgs end with many different things
        if package.startswith(("linux-tools-", "linux-firmware-")):
            return False

        if package in {
            "linux-timemachine",
            "linux-headers",
            "linux-apfs-rw-src",
        }:
            return False

        return True



# SPDX-License-Identifier: LGPL-2.1-or-later

from collections.abc import Iterable
from pathlib import Path
from typing import Any, Optional

from mkosi.context import Context
from mkosi.distribution import Distribution, centos, join_mirror
from mkosi.installer.rpm import RpmRepository, find_rpm_gpgkey
from mkosi.log import die
from mkosi.run import exists_in_sandbox, glob_in_sandbox


class Installer(centos.Installer, distribution=Distribution.rhel):
    @classmethod
    def pretty_name(cls) -> str:
        return "RHEL"

    @classmethod
    def gpgurls(cls, context: Context) -> tuple[str, ...]:
        return (
            find_rpm_gpgkey(
                context,
                f"RPM-GPG-KEY-redhat{cls.major_release(context.config)}-release",
                "https://access.redhat.com/security/data/fd431d51.txt",
            ),
        )

    @staticmethod
    def sslcacert(context: Context) -> Optional[Path]:
        if context.config.mirror:
            return None

        path = Path("/etc/rhsm/ca/redhat-uep.pem")
        if not exists_in_sandbox(path, sandbox=context.sandbox()):
            die(
                f"redhat-uep.pem certificate not found in sandbox at {path}",
                hint="Add the certificate to the sandbox with SandboxTrees= or mkosi.sandbox/",
            )

        return path

    @staticmethod
    def sslclientkey(context: Context) -> Optional[Path]:
        if context.config.mirror:
            return None

        glob = "/etc/pki/entitlement/*-key.pem"
        paths = glob_in_sandbox(glob, sandbox=context.sandbox())
        if not paths:
            die(
                f"No entitlement keys found at {glob} in sandbox",
                hint="Add an entitlement key to the sandbox with SandboxTrees= or mkosi.sandbox/",
            )

        return paths[0]

    @staticmethod
    def sslclientcert(context: Context) -> Optional[Path]:
        if context.config.mirror:
            return None

        glob = "/etc/pki/entitlement/*.pem"
        paths = [p for p in glob_in_sandbox(glob, sandbox=context.sandbox()) if "-key.pem" not in p.name]
        if not paths:
            die(
                f"No entitlement certificates found at {glob} in sandbox",
                hint="Add an entitlement certificate to the sandbox with SandboxTrees= or mkosi.sandbox/",
            )

        return paths[0]

    @classmethod
    def repository_variants(
        cls,
        context: Context,
        gpgurls: tuple[str, ...],
        repo: str,
    ) -> Iterable[RpmRepository]:
        if context.config.local_mirror:
            yield RpmRepository(repo, f"baseurl={context.config.local_mirror}", gpgurls)
        else:
            mirror = context.config.mirror or "https://cdn.redhat.com/content/dist/"

            common: dict[str, Any] = dict(
                gpgurls=gpgurls,
                sslcacert=cls.sslcacert(context),
                sslclientcert=cls.sslclientcert(context),
                sslclientkey=cls.sslclientkey(context),
            )

            v = context.config.release
            major = cls.major_release(context.config)
            yield RpmRepository(
                f"rhel-{v}-{repo}-rpms",
                f"baseurl={join_mirror(mirror, f'rhel{major}/{v}/$basearch/{repo}/os')}",
                enabled=True,
                **common,
            )
            yield RpmRepository(
                f"rhel-{v}-{repo}-debug-rpms",
                f"baseurl={join_mirror(mirror, f'rhel{major}/{v}/$basearch/{repo}/debug')}",
                enabled=False,
                **common,
            )
            yield RpmRepository(
                f"rhel-{v}-{repo}-source",
                f"baseurl={join_mirror(mirror, f'rhel{major}/{v}/$basearch/{repo}/source')}",
                enabled=False,
                **common,
            )

    @classmethod
    def repositories(cls, context: Context) -> Iterable[RpmRepository]:
        if context.config.snapshot:
            die(f"Snapshot= is not supported for {cls.pretty_name()}")

        gpgurls = cls.gpgurls(context)
        yield from cls.repository_variants(context, gpgurls, "baseos")
        yield from cls.repository_variants(context, gpgurls, "appstream")
        yield from cls.repository_variants(context, gpgurls, "codeready-builder")
        yield from cls.epel_repositories(context)



# SPDX-License-Identifier: LGPL-2.1-or-later

import fnmatch
import itertools
import logging
import os
import re
import subprocess
from collections.abc import Iterable, Iterator, Reversible
from pathlib import Path

from mkosi.context import Context
from mkosi.log import complete_step, log_step
from mkosi.run import chroot_cmd, run
from mkosi.sandbox import chase
from mkosi.util import chdir, parents_below


def loaded_modules() -> list[str]:
    # Loaded modules are listed with underscores but the filenames might use dashes instead.
    return [
        normalize_module_name(line.split()[0]) for line in Path("/proc/modules").read_text().splitlines()
    ]


def globs_match_filename(
    name: str,
    globs: Reversible[str],
    *,
    match_default: bool = False,
) -> bool:
    # Check whether the path matches any of the globs

    for glob in reversed(globs):
        # Patterns are evaluated in order and last matching one wins.
        # Patterns may be prefixed with '-' to exclude modules.
        if negative := glob.startswith("-"):
            glob = glob[1:]
        # As a special case, if a directory is specified, all items
        # below that directory are matched.
        if glob.endswith("/"):
            glob += "*"

        if (
            # Match globs starting with / relative to kernel/ first, since in-tree module are the common case
            (glob.startswith("/") and fnmatch.fnmatch(f"/{name}", f"/kernel{glob}"))
            # Now match absolute globs relative to lib/modules/KVER/
            or (glob.startswith("/") and fnmatch.fnmatch(f"/{name}", glob))
            # match a subset of the path, at path element boundary
            or ("/" in glob and fnmatch.fnmatch(f"/{name}", f"*/{glob}"))
            # match the basename
            or fnmatch.fnmatch(name.split("/")[-1], glob)
        ):
            return not negative

    return match_default


def globs_match_module(
    name: str,
    globs: Reversible[str],
) -> bool:
    # Strip '.ko' suffix and an optional compression suffix
    name = re.sub(r"\.ko(\.(gz|xz|zst))?$", "", name)
    # Check whether the suffixless-path matches any of the globs
    return globs_match_filename(name, globs)


def globs_match_firmware(
    name: str,
    globs: Reversible[str],
    *,
    match_default: bool = False,
) -> bool:
    # Strip any known compression suffixes
    name = re.sub(r"\.(gz|xz|zst)$", "", name)
    # Check whether the suffixless-path matches any of the globs
    return globs_match_filename(name, globs, match_default=match_default)


def filter_kernel_modules(
    root: Path,
    kver: str,
    *,
    include: Iterable[str],
    exclude: Iterable[str],
) -> list[str]:
    log_step("Applying kernel modules include/exclude configuration")
    if include:
        logging.debug(f"Kernel modules include directives: {' '.join(include)}")
    if exclude:
        logging.debug(f"Kernel modules exclude directives: {' '.join(exclude)}")

    modulesd = Path("usr/lib/modules") / kver
    with chdir(root):
        # The glob may match additional paths.
        # Narrow this down to *.ko, *.ko.gz, *.ko.xz, *.ko.zst.
        modules = {
            m for m in modulesd.rglob("*.ko*") if m.name.endswith((".ko", ".ko.gz", ".ko.xz", ".ko.zst"))
        }

    n_modules = len(modules)

    keep = set()

    if include:
        patterns = [p[3:] for p in include if p.startswith("re:")]
        regex = re.compile("|".join(patterns))

        globs = [normalize_module_glob(p) for p in include if not p.startswith("re:")]

        for m in modules:
            rel = os.fspath(m.relative_to(modulesd))
            # old regexes match relative to modulesd/subdir/ not modulesd/
            legacy_rel = os.fspath(Path(*m.parts[5:]))

            if (patterns and regex.search(legacy_rel)) or globs_match_module(
                normalize_module_name(rel), globs
            ):
                keep.add(rel)

    if exclude:
        assert all(p.startswith("re:") for p in exclude)
        patterns = [p[3:] for p in exclude]
        regex = re.compile("|".join(patterns))

        remove = set()
        for m in modules:
            rel = os.fspath(m.relative_to(modulesd))
            # old regexes match relative to modulesd/subdir/ not modulesd/
            legacy_rel = os.fspath(Path(*m.parts[5:]))

            if rel not in keep and regex.search(legacy_rel):
                remove.add(m)

        modules -= remove
    elif include:
        # If no exclude patterns are specified, only keep the specified kernel modules.
        modules = {modulesd / m for m in keep}

    logging.debug(f"Passing {len(modules)}/{n_modules} kernel modules on to dependency resolution.")

    return sorted(module_path_to_name(m) for m in modules)


def filter_firmware(
    root: Path,
    firmware: set[Path],
    *,
    include: Iterable[str],
    exclude: Iterable[str],
) -> set[Path]:
    log_step("Applying firmware include/exclude configuration")
    if include:
        logging.debug(f"Firmware include directives: {' '.join(include)}")
    if exclude:
        logging.debug(f"Firmware exclude directives: {' '.join(exclude)}")

    firmwared = Path("usr/lib/firmware")

    # globs can be also used to exclude firmware, so we we need to apply them
    # to the inherited list of firmware files too.
    globs = [p for p in include if not p.startswith("re:")]

    if exclude or globs:
        assert all(p.startswith("re:") for p in exclude)
        remove = set()
        patterns = [p[3:] for p in exclude]
        regex = re.compile("|".join(patterns))

        for f in firmware:
            rel = os.fspath(f.relative_to(firmwared))
            if (patterns and regex.search(rel)) or not globs_match_firmware(rel, globs, match_default=True):
                remove.add(f)

        firmware -= remove

    if include:
        with chdir(root):
            all_firmware = {p for p in firmwared.rglob("*") if p.is_file() or p.is_symlink()}

        patterns = [p[3:] for p in include if p.startswith("re:")]
        regex = re.compile("|".join(patterns))

        for f in all_firmware:
            rel = os.fspath(f.relative_to(firmwared))
            if (patterns and regex.search(rel)) or globs_match_firmware(rel, globs):
                firmware.add(f)

    logging.debug(f"A total of {len(firmware)} firmware files will be included in the image")

    return firmware


def normalize_module_name(name: str) -> str:
    # Replace '_' by '-'
    return name.replace("_", "-")


def normalize_module_glob(name: str) -> str:
    # We want to replace '_' by '-', except when used in […]
    ans = ""
    while name:
        i = (name + "[").index("[")
        ans += name[:i].replace("_", "-")
        name = name[i:]
        i = (name + "]").index("]")
        ans += name[: i + 1]
        name = name[i + 1 :]
    return ans


def module_path_to_name(path: Path) -> str:
    return normalize_module_name(path.name.partition(".")[0])


def modinfo(context: Context, kver: str, modules: Iterable[str]) -> str:
    cmdline = ["modinfo", "--set-version", kver, "--null"]

    if context.config.output_format.is_extension_image() and not context.config.overlay:
        cmdline += ["--basedir", "/buildroot"]
        sandbox = context.sandbox(options=context.rootoptions(readonly=True))
    else:
        sandbox = chroot_cmd(root=context.rootoptions)

    cmdline += [*modules]

    return run(
        cmdline,
        stdout=subprocess.PIPE,
        sandbox=sandbox,
    ).stdout.strip()


def resolve_module_dependencies(
    context: Context,
    kver: str,
    modules: Iterable[str],
) -> tuple[set[Path], set[Path]]:
    """
    Returns a tuple of lists containing the paths to the module and firmware dependencies of the given list
    of module names (including the given module paths themselves). The paths are returned relative to the
    root directory.
    """
    modulesd = Path("usr/lib/modules") / kver
    if (p := context.root / modulesd / "modules.builtin").exists():
        builtin = set(module_path_to_name(Path(m)) for m in p.read_text().splitlines())
    else:
        builtin = set()
    with chdir(context.root):
        allmodules = set(modulesd.rglob("*.ko*"))
    nametofile = {module_path_to_name(m): m for m in allmodules}

    log_step("Running modinfo to fetch kernel module dependencies")

    # We could run modinfo once for each module but that's slow. Luckily we can pass multiple modules to
    # modinfo and it'll process them all in a single go. We get the modinfo for all modules to build two maps
    # that map the path of the module to its module dependencies and its firmware dependencies
    # respectively. Because there's more kernel modules than the max number of accepted CLI arguments, we
    # split the modules list up into chunks.
    info = ""
    for i in range(0, len(nametofile.keys()), 8500):
        chunk = list(nametofile.keys())[i : i + 8500]
        info += modinfo(context, kver, chunk)

    log_step("Calculating required kernel modules and firmware")

    moddep = {}
    firmwaredep = {}

    depends: set[str] = set()
    firmware: set[Path] = set()

    with chdir(context.root):
        for line in info.split("\0"):
            key, sep, value = line.partition(":")
            if not sep:
                key, sep, value = line.partition("=")

            value = value.strip()

            if key == "depends":
                depends.update(normalize_module_name(d) for d in value.split(",") if d)

            elif key == "softdep":
                # softdep is delimited by spaces and can contain strings like pre: and post: so discard
                # anything that ends with a colon.
                depends.update(normalize_module_name(d) for d in value.split() if not d.endswith(":"))

            elif key == "firmware":
                if (Path("usr/lib/firmware") / value).exists():
                    firmware.add(Path("usr/lib/firmware") / value)

                glob = "" if value.endswith("*") else ".*"

                firmware.update(Path("usr/lib/firmware").glob(f"{value}{glob}"))

            elif key == "name":
                # The file names use dashes, but the module names use underscores. We track the names in
                # terms of the file names, since the depends use dashes and therefore filenames as well.
                name = normalize_module_name(value)

                moddep[name] = depends
                firmwaredep[name] = firmware

                depends = set()
                firmware = set()

    todo = [*builtin, *modules]
    mods = set()
    firmware = set()

    while todo:
        m = todo.pop()
        if m in mods:
            continue

        depends = moddep.get(m, set())
        for d in depends:
            if d not in nametofile and d not in builtin:
                logging.warning(f"{d} is a dependency of {m} but is not installed, ignoring ")

        mods.add(m)
        todo += depends
        firmware.update(firmwaredep.get(m, set()))

    return set(nametofile[m] for m in mods if m in nametofile), set(firmware)


def gen_required_kernel_modules(
    context: Context,
    kver: str,
    *,
    modules_include: Iterable[str],
    modules_exclude: Iterable[str],
    firmware_include: Iterable[str],
    firmware_exclude: Iterable[str],
) -> Iterator[Path]:
    modulesd = Path("usr/lib/modules") / kver
    firmwared = Path("usr/lib/firmware")

    # There is firmware in /usr/lib/firmware that is not depended on by any modules so if any firmware was
    # installed we have to take the slow path to make sure we don't copy firmware into the initrd that is not
    # depended on by any kernel modules.
    if modules_include or modules_exclude or any((context.root / firmwared).glob("*")):
        modules, firmware = resolve_module_dependencies(
            context,
            kver,
            modules=filter_kernel_modules(
                context.root,
                kver,
                include=modules_include,
                exclude=modules_exclude,
            ),
        )
    else:
        logging.debug(
            "No modules excluded and no firmware installed, using kernel modules generation fast path"
        )
        with chdir(context.root):
            modules = set(modulesd.rglob("*.ko*"))
        firmware = set()

    # Include or exclude firmware explicitly configured
    firmware = filter_firmware(context.root, firmware, include=firmware_include, exclude=firmware_exclude)

    # /usr/lib/firmware makes use of symbolic links so we have to make sure the symlinks and their targets
    # are all included.
    fwcopy = firmware.copy()
    firmware.clear()
    for fw in fwcopy:
        # Every path component from /usr/lib/firmware up to and including the firmware file itself might be a
        # symlink. We need to make sure we include all of them so we iterate over them and keep resolving
        # each symlink separately (and recursively) and add all of them to the list of firmware to add.
        #
        # As of the time of writing this logic, the only firmware that actually requires intermediate path
        # symlink resolution are the following:
        #
        # $ find /usr/lib/firmware -type l | grep -v "\."
        # /usr/lib/firmware/intel/sof-ace-tplg
        # /usr/lib/firmware/nvidia/ad103
        # /usr/lib/firmware/nvidia/ad104
        # /usr/lib/firmware/nvidia/ad106
        # /usr/lib/firmware/nvidia/ad107
        # /usr/lib/firmware/nvidia/ga103/gsp
        # /usr/lib/firmware/nvidia/ga104/gsp
        # /usr/lib/firmware/nvidia/ga106/gsp
        # /usr/lib/firmware/nvidia/ga107/gsp
        # /usr/lib/firmware/nvidia/gb102
        # /usr/lib/firmware/nvidia/gb203
        # /usr/lib/firmware/nvidia/gb205
        # /usr/lib/firmware/nvidia/gb206
        # /usr/lib/firmware/nvidia/gb207
        # /usr/lib/firmware/nvidia/tu104/gsp
        # /usr/lib/firmware/nvidia/tu106/gsp
        # /usr/lib/firmware/nvidia/tu117/gsp

        todo = list(reversed(fw.parts))
        current = context.root
        while todo:
            part = todo.pop()
            if part == "/":
                current = context.root
                continue
            elif part == "..":
                current = current.parent
                continue
            elif part == ".":
                continue

            current /= part
            if not current.is_symlink():
                continue

            if current.readlink().is_relative_to("/etc/alternatives"):
                target = chase(os.fspath(context.root), os.fspath(current.relative_to(context.root)))
                current.unlink()
                current.symlink_to(os.path.relpath(target, start=current.parent))

            firmware.add(current.relative_to(context.root))
            todo += list(reversed(current.readlink().parts))
            # Relative symlinks are resolved relative to the directory
            # the symlink is located in. If the symlink is absolute we'll
            # override the current path anyway so modifying it here doesn't
            # matter.
            current = current.parent

        # Finally, add the actual fully resolved path to the firmware file.
        if current.exists():
            firmware.add(current.relative_to(context.root))

    yield from sorted(
        itertools.chain(
            {
                p.relative_to(context.root)
                for f in modules | firmware
                for p in parents_below(context.root / f, context.root / "usr/lib")
            },
            modules,
            firmware,
            (p.relative_to(context.root) for p in (context.root / modulesd).glob("modules*")),
        )
    )

    if (modulesd / "vdso").exists():
        if not modules:
            yield from (
                p.relative_to(context.root)
                for p in parents_below(context.root / modulesd / "vdso", context.root / "usr/lib")
            )

        yield modulesd / "vdso"
        yield from sorted(p.relative_to(context.root) for p in (context.root / modulesd / "vdso").iterdir())


def process_kernel_modules(
    context: Context,
    kver: str,
    *,
    modules_include: Iterable[str],
    modules_exclude: Iterable[str],
    firmware_include: Iterable[str],
    firmware_exclude: Iterable[str],
) -> None:
    if not (modules_include or modules_exclude or firmware_include or firmware_exclude):
        return

    modulesd = Path("usr/lib/modules") / kver
    firmwared = Path("usr/lib/firmware")

    with complete_step("Applying kernel module filters"):
        required = set(
            gen_required_kernel_modules(
                context,
                kver,
                modules_include=modules_include,
                modules_exclude=modules_exclude,
                firmware_include=firmware_include,
                firmware_exclude=firmware_exclude,
            )
        )

        with chdir(context.root):
            modules = sorted(modulesd.rglob("*.ko*"), reverse=True)
            firmware = sorted(firmwared.rglob("*"), reverse=True)

        for m in modules:
            if m in required:
                continue

            p = context.root / m
            if p.is_file() or p.is_symlink():
                if p.is_symlink():
                    p_target = Path(chase(os.fspath(context.root), os.fspath(m)))
                    if p_target.exists():
                        p_target.unlink()
                p.unlink()
            elif p.exists():
                p.rmdir()

        for fw in firmware:
            if fw in required:
                continue

            if any(fw.is_relative_to(firmwared / d) for d in ("amd-ucode", "intel-ucode")):
                continue

            p = context.root / fw
            if p.is_file() or p.is_symlink():
                p.unlink()
                if p.parent != context.root / firmwared and not any(p.parent.iterdir()):
                    p.parent.rmdir()
            elif p.exists():
                p.rmdir()


def is_valid_kdir(kdir: Path) -> bool:
    dircontent = list(kdir.glob("*"))

    # kdir does not exist or is empty
    if not dircontent:
        return False

    # check that kdir contains more than just updates
    return dircontent != [kdir / "updates"]


def filter_devicetrees(
    root: Path,
    kver: str,
    *,
    include: Iterable[str],
) -> list[Path]:
    if not include:
        return []

    logging.debug(f"Devicetrees include: {' '.join(include)}")

    # Search standard DTB locations
    dtb_dirs = [
        Path("usr/lib/firmware") / kver / "device-tree",
        Path(f"usr/lib/linux-image-{kver}"),
        Path("usr/lib/modules") / kver / "dtb",
    ]

    matched_dtbs = []
    globs = list(include)

    with chdir(root):
        for dtb_dir in dtb_dirs:
            all_dtbs = [p for p in dtb_dir.rglob("*.dtb") if p.is_file() or p.is_symlink()]
            logging.debug(f"Found {len(all_dtbs)} DTB files in {dtb_dir}")

            for dtb in all_dtbs:
                rel_path = os.fspath(dtb.relative_to(dtb_dir))
                if globs_match_filename(rel_path, globs):
                    logging.debug(f"Matched DTB: {rel_path} in {dtb_dir}")
                    matched_dtbs.append(dtb)

    if not matched_dtbs:
        logging.warning(f"Devicetrees patterns '{globs}' matched 0 files")
    else:
        logging.debug(f"Including {len(matched_dtbs)} devicetree files")

    return sorted(matched_dtbs)



# SPDX-License-Identifier: LGPL-2.1-or-later

import os
import subprocess
from pathlib import Path
from typing import Optional, overload

from mkosi.config import Config
from mkosi.mounts import finalize_certificate_mounts
from mkosi.run import run, workdir


@overload
def curl(
    config: Config,
    url: str,
    *,
    output_dir: Optional[Path],
    log: bool = True,
) -> None: ...


@overload
def curl(
    config: Config,
    url: str,
    *,
    output_dir: None = None,
    log: bool = True,
) -> str: ...


def curl(config: Config, url: str, *, output_dir: Optional[Path] = None, log: bool = True) -> Optional[str]:
    result = run(
        [
            "curl",
            "--location",
            *(["--output-dir", workdir(output_dir)] if output_dir else []),
            *(["--remote-name"] if output_dir else []),
            "--no-progress-meter",
            "--fail",
            *(["--silent"] if not log else []),
            *(["--proxy", config.proxy_url] if config.proxy_url else []),
            *(["--noproxy", ",".join(config.proxy_exclude)] if config.proxy_exclude else []),
            *(["--proxy-capath", "/proxy.cacert"] if config.proxy_peer_certificate else []),
            *(["--proxy-cert", "/proxy.clientcert"] if config.proxy_client_certificate else []),
            *(["--proxy-key", "/proxy.clientkey"] if config.proxy_client_key else []),
            url,
        ],
        stdout=None if output_dir else subprocess.PIPE,
        sandbox=config.sandbox(
            network=True,
            options=[
                *(["--bind", os.fspath(output_dir), workdir(output_dir)] if output_dir else []),
                *finalize_certificate_mounts(config)
            ],
        ),
        log=log,
    )  # fmt: skip

    return None if output_dir else result.stdout



# SPDX-License-Identifier: LGPL-2.1-or-later

from collections.abc import Iterable

from mkosi.config import Architecture
from mkosi.context import Context
from mkosi.distribution import (
    Distribution,
    fedora,
    join_mirror,
)
from mkosi.installer.dnf import Dnf
from mkosi.installer.rpm import RpmRepository, find_rpm_gpgkey, setup_rpm
from mkosi.log import die


class Installer(fedora.Installer, distribution=Distribution.azure):
    @classmethod
    def pretty_name(cls) -> str:
        return "Azure Linux"

    @classmethod
    def default_release(cls) -> str:
        return "3.0"

    @classmethod
    def filesystem(cls) -> str:
        return "ext4"

    @classmethod
    def setup(cls, context: Context) -> None:
        setup_rpm(context, dbpath="/var/lib/rpm")
        Dnf.setup(context, list(cls.repositories(context)), filelists=False)

    @classmethod
    def install(cls, context: Context) -> None:
        cls.install_packages(context, ["filesystem", "azurelinux-release"], apivfs=False)

    @classmethod
    def repositories(cls, context: Context) -> Iterable[RpmRepository]:
        if context.config.snapshot:
            die(f"Snapshot= is not supported for {cls.pretty_name()}")

        gpgurls = (
            find_rpm_gpgkey(
                context,
                "MICROSOFT-RPM-GPG-KEY",
                "https://raw.githubusercontent.com/rpm-software-management/distribution-gpg-keys/main/keys/azure-linux/MICROSOFT-RPM-GPG-KEY",
            ),
        )

        if context.config.local_mirror:
            yield RpmRepository("base", f"baseurl={context.config.local_mirror}", gpgurls)
            return

        mirror = context.config.mirror or "https://packages.microsoft.com/azurelinux"
        url = join_mirror(mirror, context.config.release)

        for repo in ("base", "extended", "ms-oss", "ms-non-oss", "cloud-native", "nvidia"):
            yield RpmRepository(
                repo,
                f"baseurl={url}/prod/{repo}/$basearch",
                gpgurls,
            )

            repo = "NVIDIA" if repo == "nvidia" else repo
            yield RpmRepository(
                f"{repo}-preview",
                f"baseurl={url}/preview/{repo}/$basearch",
                gpgurls,
                enabled=False,
            )

        for repo in ("base", "cloud-native", "extended"):
            yield RpmRepository(
                f"{repo}-debuginfo",
                f"baseurl={url}/prod/{repo}/debuginfo/$basearch",
                gpgurls,
                enabled=False,
            )
            yield RpmRepository(
                f"{repo}-preview-debuginfo",
                f"baseurl={url}/preview/{repo}/debuginfo/$basearch",
                gpgurls,
                enabled=False,
            )

        for repo in ("base", "cloud-native", "extended", "ms-oss"):
            yield RpmRepository(
                f"{repo}-source",
                f"baseurl={url}/prod/{repo}/srpms",
                gpgurls,
                enabled=False,
            )
            yield RpmRepository(
                f"{repo}-source",
                f"baseurl={url}/preview/{repo}/srpms",
                gpgurls,
                enabled=False,
            )

    @classmethod
    def architecture(cls, arch: Architecture) -> str:
        a = {
            Architecture.arm64:  "aarch64",
            Architecture.x86_64: "x86_64",
        }.get(arch)  # fmt: skip

        if not a:
            die(f"Architecture {a} is not supported by {cls.pretty_name()}")

        return a



# SPDX-License-Identifier: PSF-2.0
# Based on code from https://github.com/python/cpython/blob/main/Lib/importlib/resources/_common.py

import contextlib
import functools
import os
import stat
import sys
import tempfile
from collections.abc import Iterator
from contextlib import AbstractContextManager
from pathlib import Path

if sys.version_info >= (3, 11):
    from importlib.resources.abc import Traversable
else:
    from importlib.abc import Traversable


def make_executable_if_script(path: Path) -> Path:
    with path.open("rb") as f:
        buf = f.read(3)
        if buf == b"#!/":
            os.chmod(path, path.stat().st_mode | stat.S_IEXEC)

    return path


@contextlib.contextmanager
def temporary_file(path: Traversable, suffix: str = "") -> Iterator[Path]:
    fd, raw_path = tempfile.mkstemp(suffix=suffix)
    try:
        try:
            os.write(fd, path.read_bytes())
        finally:
            os.close(fd)
        yield make_executable_if_script(Path(raw_path))
    finally:
        try:
            os.remove(raw_path)
        except FileNotFoundError:
            pass


def dir_is_present(path: Traversable) -> bool:
    """
    Some Traversables implement ``is_dir()`` to raise an
    exception (i.e. ``FileNotFoundError``) when the
    directory doesn't exist. This function wraps that call
    to always return a boolean and only return True
    if there's a dir and it exists.
    """
    with contextlib.suppress(FileNotFoundError):
        return path.is_dir()
    return False


@functools.singledispatch
def as_file(path: Traversable) -> AbstractContextManager[Path]:
    """
    Given a Traversable object, return that object as a
    path on the local file system in a context manager.
    """
    return temporary_dir(path) if dir_is_present(path) else temporary_file(path, suffix=path.name)


@contextlib.contextmanager
def temporary_dir(path: Traversable) -> Iterator[Path]:
    """
    Given a traversable dir, recursively replicate the whole tree
    to the file system in a context manager.
    """
    assert path.is_dir()
    with tempfile.TemporaryDirectory() as temp_dir:
        yield write_contents(Path(temp_dir), path)


def write_contents(target: Path, source: Traversable) -> Path:
    child = target.joinpath(source.name)
    if source.is_dir():
        child.mkdir()
        for item in source.iterdir():
            write_contents(child, item)
    else:
        child.write_bytes(source.read_bytes())
        make_executable_if_script(child)
    return child
